/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "DSMC.h"
#include "EnergyModelC.h"
#include "SearchSpaceC.h"
#include "DSMCallbackI.h"

#include <meshlib/MeshC.h>
#include <meshlib/MexelC.h>

#include <deque>

// added 17.9. by J.M.
#include <blitz/tinyvec-et.h>

using namespace dsmlib;
using namespace std;
using namespace meshlib;

DSMC::ParametersC::ParametersC() throw ()
{
}

DSMC::ParametersC::ParametersC(
	meshlib::MLFloat lambda,
	unsigned int searchJ,
	unsigned int searchK,
	meshlib::MLFloat searchL,
	meshlib::MLFloat searchD,
	meshlib::MLFloat tolerance,
	meshlib::MLFloat volRatio,
	meshlib::MLFloat gamma ) throw ()
:
	m_lambda(lambda),
	m_searchJ(searchJ),
	m_searchK(searchK),
	m_searchL(searchL),
	m_searchD(searchD),
	m_tolerance(tolerance),
	m_volRatio(volRatio),
	m_gamma(gamma)
{
}

DSMC::ParametersC::~ParametersC() throw ()
{
}

DSMC::ResultsC::ResultsC(
	meshlib::MeshC * pOuter,
	meshlib::MLFloat outerEnergy,
	meshlib::MeshC * pInner,
	meshlib::MLFloat innerEnergy,
	int searchDirection,
	unsigned int rounds,
	const DSMC::ParametersC & rParams ) throw ()
:
	m_pOuter(pOuter),
	m_outerEnergy(outerEnergy),
	m_pInner(pInner),
	m_innerEnergy(innerEnergy),
	m_searchDirection(searchDirection),
	m_rounds(rounds),
	m_rParams(rParams)
{
}

DSMC::ResultsC::~ResultsC() throw ()
{
}

meshlib::MeshC * DSMC::ResultsC::getMesh() const throw ()
{
	if ( m_searchDirection == 0 )
		return (m_outerEnergy < m_innerEnergy) ? m_pOuter : m_pInner;
	else
		return ( m_searchDirection == -1) ? m_pOuter : m_pInner;
}

meshlib::MeshC * DSMC::ResultsC::getOuterMesh() const throw ()
{
	return m_pOuter;
}

meshlib::MeshC * DSMC::ResultsC::getInnerMesh() const throw ()
{
	return m_pInner;
}

meshlib::MLFloat DSMC::ResultsC::getEnergy() const throw ()
{
	return (m_outerEnergy < m_innerEnergy) ? m_outerEnergy : m_innerEnergy;
}

meshlib::MLFloat DSMC::ResultsC::getOuterEnergy() const throw ()
{
	return m_outerEnergy;
}

meshlib::MLFloat DSMC::ResultsC::getInnerEnergy() const throw ()
{
	return m_innerEnergy;
}

unsigned int DSMC::ResultsC::getRounds() const throw ()
{
	return m_rounds;
}

const DSMC::ParametersC & DSMC::ResultsC::getParameters() const throw ()
{
	return m_rParams;
}

int DSMC::ResultsC::getSearchDirection() const throw ()
{
	return m_searchDirection;
}

DSMC::DSMC(
	meshlib::MeshC * pOuter,
	meshlib::MeshC * pInner,
	const imagelib::ImageI * pImage ) throw ()
:
	m_pCallback(NULL),
	m_pOuter(pOuter),
	m_pInner(pInner),
	m_pImage(pImage),
	m_oscRounds(100),
	m_oscTolerance(0.97),
	mexLenMin(3.0)
{	
}

DSMC::~DSMC() throw ()
{
}

///A helper function.
template <class T>
bool isNaN( T v )
{
    return !(0.0 <= v) && !(v <= 0.0);
}


DSMC::ResultsC DSMC::minimize(
	const ParametersC * pParams,
	unsigned int maxRounds,
	int searchDirection)
{
	// a queue to store inner volumes
	deque<MLFloat> innerVolumes;

	// a queue to store outer volumes
	deque<MLFloat> outerVolumes;	

	bool forcedMovement = false;
	
//	cout << "minimize() called" << endl;
	
	assert( pParams != NULL );
	
	m_pParams = pParams;
	
	MLFloat outerEnergy = numeric_limits<MLFloat>::max();
	MLFloat innerEnergy = numeric_limits<MLFloat>::max();
	MLFloat outerVolume = 0.0;
	MLFloat innerVolume = 0.0;
	
	unsigned int round;
	if ( searchDirection == -1 ) // DSM-OS
	{		
		assert( m_pOuter != NULL );
		
		m_pOuter->calcAverageArea();
		outerEnergy = getMeshEnergy( m_pOuter );
		outerVolume = m_pOuter->getVolumeMetric();

		MLFloat originalVolume = outerVolume;
		vector<MLVector> minLocs;

		m_pOuter->storeMexelLocations(minLocs);

		MLFloat minEnergy = outerEnergy;
		round = 0;
		while (
			round < maxRounds &&
			outerVolume > originalVolume * m_pParams->m_volRatio)
		{
			outerEnergy = minimizeMesh(
				m_pOuter, searchDirection, 0.0, forcedMovement );

			int r = 1;
			while ( ! m_pOuter->hasMoved() && ! isNaN(outerEnergy) )
			{
			  //cout << "O in local minimum!" << endl;
				outerEnergy = minimizeMesh(
					m_pOuter,
					searchDirection,
					m_pParams->m_gamma*r,
					forcedMovement );
				++r;
			}
			outerVolume = m_pOuter->getVolumeMetric();

			// oscillation-detection specific code				
			outerVolumes.push_back(outerVolume);
			if ( outerVolumes.size() > m_oscRounds )
				outerVolumes.pop_front();
				
			if ( forcedMovement )
			{
				if ( outerVolumes.size() == m_oscRounds &&
					 outerVolume < m_oscTolerance * (*outerVolumes.begin()) )
				{
				  //cout << "unforced" << endl;
					 forcedMovement = false;
					 outerVolumes.clear();
				}
			}
			else
			{
				if ( outerVolumes.size() == m_oscRounds &&
					 outerVolume > m_oscTolerance * (*outerVolumes.begin()) )
				{
				  //cout << "forced" << endl;
					forcedMovement = true;
				}
			}
			
			if ( outerEnergy < minEnergy )
			{
			  // cout << "Minimal ";
				m_pOuter->storeMexelLocations(minLocs);
				minEnergy = outerEnergy;
				
			}
			m_pOuter->calcAverageArea();
			++round;
			
			if ( m_pCallback )
				m_pCallback->handle( ResultsC(
					m_pOuter, outerEnergy,
					m_pInner, innerEnergy,
					searchDirection,
					round,
					*m_pParams));

//			cout << round << "\to: " << m_pOuter->getAverageArea()
//				<< ", " << outerVolume << ", " << outerEnergy << endl;
		}
		m_pOuter->restoreMexelLocations(minLocs);
		m_pOuter->calcAverageArea();
		outerVolume = m_pOuter->getVolumeMetric();
		outerEnergy = minEnergy;
	}
	else if ( searchDirection == 1 ) // DSM-IS
	{		
		assert( m_pInner != NULL );
		
		m_pInner->calcAverageArea();
		innerEnergy = getMeshEnergy( m_pInner );
		innerVolume = m_pInner->getVolumeMetric();
		MLFloat originalVolume = innerVolume;
		vector<MLVector> minLocs;
		m_pInner->storeMexelLocations(minLocs);
		MLFloat minEnergy = innerEnergy;
		round = 0;
		while (
			round < maxRounds &&
			innerVolume < originalVolume / m_pParams->m_volRatio)
		{
			innerEnergy = minimizeMesh(
				m_pInner, searchDirection, 0.0, forcedMovement );
			int r = 1;
			while ( ! m_pInner->hasMoved() && ! isNaN(outerEnergy))
			{
//				cout << "I in local minimum!" << endl;
				innerEnergy = minimizeMesh(
					m_pInner,
					searchDirection,
					m_pParams->m_gamma*r,
					forcedMovement );
				++r;
			}
			innerVolume = m_pInner->getVolumeMetric();

			// oscillation-detection specific code				
			innerVolumes.push_back(innerVolume);
			if (innerVolumes.size() > m_oscRounds )
				innerVolumes.pop_front();
			if ( forcedMovement )
			{
				if ( innerVolumes.size() == m_oscRounds &&
					 innerVolume * m_oscTolerance > (*innerVolumes.begin()) )
				{
					 forcedMovement = false;
					 innerVolumes.clear();
				}
			}
			else
			{
				if ( innerVolumes.size() == m_oscRounds &&
					 innerVolume * m_oscTolerance < (*innerVolumes.begin()) )
					 forcedMovement = true;
			}
			
			if ( innerEnergy < minEnergy )
			{
//				cout << "Minimal ";
				m_pInner->storeMexelLocations(minLocs);
				minEnergy = innerEnergy;
				
			}
			m_pInner->calcAverageArea();
			++round;

			if ( m_pCallback )
				m_pCallback->handle( ResultsC(
					m_pOuter, outerEnergy,
					m_pInner, innerEnergy,
					searchDirection,
					round,
					*m_pParams));

//			cout << round << "\ti: " << m_pInner->getAverageArea()
//				<< ", " << innerVolume << ", " << innerEnergy << endl;
		}
		m_pInner->restoreMexelLocations(minLocs);
		m_pInner->calcAverageArea();
		innerVolume = m_pInner->getVolumeMetric();
		innerEnergy = minEnergy;
	}
	else if ( searchDirection == 0 ) // DSM
	{
	
		m_pOuter->calcAverageArea();
		m_pInner->calcAverageArea();
		outerEnergy = getMeshEnergy( m_pOuter );
		innerEnergy = getMeshEnergy( m_pInner );
		outerVolume = m_pOuter->getVolumeMetric();
		innerVolume = m_pInner->getVolumeMetric();
		round = 0;
		
		while (
			round < maxRounds &&
			outerVolume > innerVolume * m_pParams->m_tolerance )
		{
			if ( outerEnergy > innerEnergy )
			{				
				outerEnergy = minimizeMesh( m_pOuter, -1, 0.0, forcedMovement );
				int r = 1;
				while ( ! m_pOuter->hasMoved() && ! isNaN(outerEnergy) )
				{
//					cout << "O in local minimum!" << endl;
					outerEnergy = minimizeMesh(
						m_pOuter, -1, m_pParams->m_gamma*r,
						forcedMovement );
					++r;
				}
				outerVolume = m_pOuter->getVolumeMetric();
				innerVolumes.clear();

				// oscillation-detection specific code				
				outerVolumes.push_back(outerVolume);
				if ( outerVolumes.size() > m_oscRounds )
					outerVolumes.pop_front();
					
				if ( forcedMovement )
				{
					if ( outerVolumes.size() == m_oscRounds &&
						 outerVolume * m_oscTolerance > (*outerVolumes.begin()) )
					{
//						 cout << "O forced movement disabled" << endl;
						 forcedMovement = false;
						 outerVolumes.clear();
					}
				}
				else
				{
					if ( outerVolumes.size() == m_oscRounds &&
						 outerVolume * m_oscTolerance < (*outerVolumes.begin()) )
					{
//						cout << " O forced movement enabled" << endl;
						forcedMovement = true;
					}
				}
				
			}
			else
			{				
				innerEnergy = minimizeMesh( m_pInner, 1, 0.0, forcedMovement );
				int r = 1;
				while ( ! m_pInner->hasMoved() && ! isNaN(outerEnergy) )
				{
//					cout << "I in local minimum!" << endl;
					innerEnergy = minimizeMesh(
						m_pInner, 1, m_pParams->m_gamma*r,
						forcedMovement );
					++r;
				}
				innerVolume = m_pInner->getVolumeMetric();
				outerVolumes.clear();

				// oscillation-detection specific code				
				innerVolumes.push_back(innerVolume);
				if ( innerVolumes.size() > m_oscTolerance )
					innerVolumes.pop_front();
					
				if ( forcedMovement )
				{
					if ( innerVolumes.size() == m_oscRounds &&
						 innerVolume * m_oscTolerance > (*innerVolumes.begin()) )
					{
// 						cout << "I forced movement disabled" << endl;
						 forcedMovement = false;
						 innerVolumes.clear();
					}
				}
				else
				{
					if ( innerVolumes.size() == m_oscRounds &&
						 innerVolume * m_oscTolerance < (*innerVolumes.begin()) )
					{
//						cout << "I forced movement enabled" << endl;
						 forcedMovement = true;
					}
				}
			}
						
			m_pOuter->calcAverageArea();
			m_pInner->calcAverageArea();
	
			++round;

			if ( m_pCallback )
				m_pCallback->handle( ResultsC(
					m_pOuter, outerEnergy,
					m_pInner, innerEnergy,
					searchDirection,
					round,
					*m_pParams));
				
		}
	}
	else
	{
		throw iolib::ExceptionC("Search direction must be -1, +1 or 0.");
	}
		
	return ResultsC(
		m_pOuter, outerEnergy,
		m_pInner, innerEnergy,
		searchDirection,
		round,
		*m_pParams);
}

MLFloat DSMC::getMeshEnergy( const MeshC * pMesh ) const throw ()
{
  MLFloat energy = 0.0;
  MexelConstIterator itMex = pMesh->mexelBegin();
  while ( itMex != pMesh->mexelEnd() )
    {
      energy += EnergyModelC::getMexelEnergy(**itMex, *pMesh, *m_pImage, m_pParams->m_lambda );
      ++itMex;
    }
  return energy;
}

meshlib::MLFloat DSMC::minimizeMesh( MeshC * pMesh, int searchDirection,
	MLFloat penalty,
	bool forcedMovement )
{
	///@todo cache mexel energies = few% speed-up.
  //if(penalty>0) cout << "penalty: " << penalty << endl;
	pMesh->setMoved(false);
	
	MLFloat energy = 0.0;
	MexelIterator itMex = pMesh->mexelBegin();
	while ( itMex != pMesh->mexelEnd() )
	{
		m_space.recalculate(
			**itMex,
			searchDirection,
			m_pParams->m_searchJ,
			// forcedMovement ? 0 : m_pParams->m_searchK, // changed 22.09
			(forcedMovement || penalty>0.0) ? 0 : m_pParams->m_searchK,
			m_pParams->m_searchL,
			m_pParams->m_searchD);

		bool mexMoved = false; // Added 20.9.
		VectorIterator itLoc = m_space.begin();
		VectorIterator itMinLoc = itLoc;
		MLFloat minEnergy = EnergyModelC::getMexelEnergy(
				**itMex, *pMesh, *m_pImage, m_pParams->m_lambda ) + penalty;

		++itLoc; // this assumes that the first location in the search space
				// is the mexel's current location.

		while ( itLoc != m_space.end() )
		{
			(*itMex)->setLocation(**itLoc);
			MLFloat mexelEnergy = EnergyModelC::getMexelEnergy(
				**itMex, *pMesh, *m_pImage, m_pParams->m_lambda );
			if ( mexelEnergy < minEnergy )
			{
				minEnergy = mexelEnergy;
				itMinLoc = itLoc;
				//  pMesh->setMoved(true);  // removed 20.9.
				mexMoved = true;  //  added 20.9.
			}
			++itLoc;
		}

		// Following added by J.M. 20.9.
		// It cancels unwanted mexel movements near the reference point.
		MLVector mexLoc( (*itMex)->getLocation() );
		MLFloat mexLen = sqrt( sum( mexLoc * mexLoc ) ); 
		//MLFloat mexLen= sqrt(mexLoc[0]*mexLoc[0] + mexLoc[1]*mexLoc[1] +  mexLoc[2]*mexLoc[2]);

		if(mexLen<mexLenMin) 
		  { 
		    itMinLoc = m_space.begin();
		    (*itMex)->setLocation(**itMinLoc);
		    minEnergy = EnergyModelC::getMexelEnergy(**itMex, *pMesh, *m_pImage, 
						 m_pParams->m_lambda ) + penalty;
		    mexMoved = false;
		  } 

		if(mexMoved) 
		  pMesh->setMoved(true);

		//  cout << mexLoc << " " <<  minEnergy <<"\n";

		// end of changes

		(*itMex)->setLocation(**itMinLoc);
		energy += minEnergy;
		++itMex;
	}
		
	return energy;
}


void DSMC::setCallback( DSMCallbackI * pCallback ) throw ()
{
	m_pCallback = pCallback;
}



