/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef DSMLIB_DSMC_H
#define DSMLIB_DSMC_H

#include "SearchSpaceC.h"

#include <meshlib/typedefs.h>

//forward declarations

namespace meshlib
{
	class MeshC;
};

namespace imagelib
{
	class ImageI;
};

/**Namespace for the DSM algorithm.
 * 
 */
namespace dsmlib
{
	// forward declarations
	class DSMCallbackI;
	
	/**DSM algorithm implementation.
	 * This class implements the DSM algorithm.
	 */
	class DSMC
	{
	public:
	
		/**Minimization parameters.
		 * This is an inner class for minimization parameters.
		 * Basically just a struct.
		 */
		class ParametersC
		{
		public:
		
			/**DSMC needs access to the default constructor and to
			 * the data members.
			 */
			friend class DSMC;
		
			/**Constructor.
			 * This is the constructor that should be used.
			 * @param lambda Regularization parameter
			 * @param searchJ Search space J
			 * @param searchK Search space K
			 * @param searchL Search space L
			 * @param searchD Search space D
			 * @param tolerance Volume tolerance
			 * @param volRatio Volume ratio for single-surface method
			 * @param gamma Penalty coefficient 
			 */
			ParametersC(
				meshlib::MLFloat lambda,
				unsigned int searchJ,
				unsigned int searchK,
				meshlib::MLFloat searchL,
				meshlib::MLFloat searchD,
				meshlib::MLFloat tolerance,
				meshlib::MLFloat volRatio,
				meshlib::MLFloat gamma ) throw ();

			/**Destructor.
			 * Note that as this class is not designed to be subclassed,
			 * the destructor is not virtual.
			 */
			~ParametersC() throw ();
			
		private:

			/**Default constructor.
			 * Needed for practical reasons.
			 */
			ParametersC() throw ();
			
			meshlib::MLFloat m_lambda;
			unsigned int m_searchJ;
			unsigned int m_searchK;
			meshlib::MLFloat m_searchL;
			meshlib::MLFloat m_searchD;
			meshlib::MLFloat m_tolerance;
			meshlib::MLFloat m_volRatio;
			meshlib::MLFloat m_gamma;

		}; // class ParametersC

		/**Minimization results.
		 *
		 */
		class ResultsC
		{
		public:	
			
			/**DSMC needs to access the constructor
			 */
			friend class DSMC;
			
			/**A non-virtual destructor.
			 * This class is not designed for subclassing.
			 */
			~ResultsC() throw ();
			
			/**Return the mesh with the lower energy.
			 * @return A pointer to the mesh with the lower energy.
			 */
			meshlib::MeshC * getMesh() const throw ();
			
			/**Return outer mesh.
			 * @return A pointer to the outer mesh
			 */
			meshlib::MeshC * getOuterMesh() const throw ();

			/**Return inner mesh.
			 * @return A pointer to the inner mesh
			 */
			meshlib::MeshC * getInnerMesh() const throw ();

			/**Return minimun energy.
			 * @return Energy of the mesh with the lower energy.
			 */
			 
			meshlib::MLFloat getEnergy() const throw ();

			/**Return outer mesh energy.
			 * @return Outer mesh energy
			 */
			meshlib::MLFloat getOuterEnergy() const throw ();

			/**Return inner mesh energy.
			 * @return Inner mesh energy
			 */
			meshlib::MLFloat getInnerEnergy() const throw ();

			/**Return the number of rounds completed.
			 * @return The number of rounds.
			 */
			unsigned int getRounds() const throw ();

			/**Return the parameter set used.
			 * @return A reference to the parameter set
			 */
			const DSMC::ParametersC & getParameters() const throw ();

			/**Return search direction.
			 * @return 0 for dual-surface algorithm; -1 for outer surface
			 * algorithm; 1 for inner surface algorith.
			 */
			int getSearchDirection() const throw ();
			
		private:
		
			/**Private constructor.
			 * Used by class DSMC.
			 * @param pOuter A pointer to the outer mesh
			 * @param outerEnergy Energy of the outer mesh
			 * @param pInner A pointer to the inner mesh
			 * @param innerEnergy Energy of the inner mesh
			 * @param searchDirection Search direction
			 * @param rounds The number of rounds completed
			 * @param rParams A reference to the parameter set
			 */
			ResultsC(
				meshlib::MeshC * pOuter,
				meshlib::MLFloat outerEnergy,
				meshlib::MeshC * pInner,
				meshlib::MLFloat innerEnergy,
				int searchDirection,
				unsigned int rounds,
				const DSMC::ParametersC & rParams				
				) throw ();
			
			///Outer mesh.
			meshlib::MeshC * m_pOuter;
			
			///Outer mesh energy.
			meshlib::MLFloat m_outerEnergy;

			//Inner mesh.
			meshlib::MeshC * m_pInner;
			
			///Inner mesh energy.
			meshlib::MLFloat m_innerEnergy;
			
			///Search direction. -1, +1 or 0.
			int m_searchDirection;

			///Number of rounds completed.
			unsigned int m_rounds;
			
			///Parameter set used.
			const DSMC::ParametersC & m_rParams;
			
		}; // class ResultsC
		
		/**A constructor.
		 * No copy is made of the input meshes;
		 * input meshes are modified by DSMC::minimize().
		 * @param pOuter A pointer to the initial outer mesh
		 * @param pInner A pointer to the initial inner mesh
		 * @param pImage A pointer to the input image.
		 */
		DSMC(
			meshlib::MeshC * pOuter,
			meshlib::MeshC * pInner,
			const imagelib::ImageI * pImage ) throw ();

		/**A virtual destructor for safety reasons.
		 */
		virtual ~DSMC() throw ();

		/**Perform the actual minimization.
		 * @param pParams A pointer to the parameter set.
		 * @param maxRounds Maximum number of rounds.
		 * @param searchDirection 0 for dsm; -1 for dsm-os; 1 for dsm-is.
		 * @return A reference to minimization results.
		 */
		ResultsC minimize(
			const ParametersC * pParams,
			unsigned int maxRounds,
			int searchDirection = 0 );
				
		/**Set callback handler.
		 * This method allows the user to provide her own callback handler
		 * to handle intermediate results.
		 * @param pCallback A pointer to the user's callback handler
		 */
		void setCallback(
			DSMCallbackI * pCallback ) throw ();
			
	private:
	
		/**Calculates the total energy of a single mesh.
		 * @param pMesh A pointer to a mesh
		 */
		meshlib::MLFloat getMeshEnergy(
			const meshlib::MeshC * pMesh ) const throw ();
			
		/**Performs one round of minimization on a single mesh.
		 * @param pMesh A pointer to a mesh to be minimized
		 * @param searchDirection -1 for the outer surface, 1 for inner
		 * @param penalty A penalty term for mexel enrgy calculations
		 */
		meshlib::MLFloat minimizeMesh(
			meshlib::MeshC * pMesh, int searchDirection,
			meshlib::MLFloat penalty,
			bool forcedMovement );

		/**Forces the mesh out of oscillating state by setting 
		 * the search space parameter K = 0.
		 * @param pMesh A pointer to a mesh to be forced
		 * @param searchDirection -1 for the outer surface, 1 for inner
		 */
/*		meshlib::MLFloat forceMesh(
			meshlib::MeshC * pMesh, int searchDirection);*/

		///Callback handler. Can be NULL. Owned by someone else.
		DSMCallbackI * m_pCallback;
		
		///Outer mesh. Can be NULL. Owned by someone else.
		meshlib::MeshC * m_pOuter;
		
		///Inner mesh. Can be NULL. Owned by someone else.
		meshlib::MeshC * m_pInner;
		
		///Input image. != NULL. Owned by someone else.
		const imagelib::ImageI * m_pImage;
		
		///Algorithm parameters. Owned by someone else.
		const ParametersC * m_pParams;
		
		SearchSpaceC m_space;

		/**the number of rounds the volume of a mesh can remain near constant
		 * until it is determined to be oscillating
		 */
		const unsigned int m_oscRounds;

		/**volume of a mesh is said to be near constant if change is less than
		 */
		const meshlib::MLFloat m_oscTolerance;

		/**minimum length for a mexel
		 */
		const meshlib::MLFloat mexLenMin;
			
	}; // class DSMC
	
}; // namespace dsmlib

#endif //DSMLIB_DSMC_H
