/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CallbackHandlerC.h"
#include "ParamDefsC.h"

#include <meshlib/MeshC.h>
#include <meshlib/OFFMeshLoaderC.h>
#include <meshlib/VRMLMeshLoaderC.h>

#include <iostream>
#include <sstream>

using namespace dsmlib;
using namespace std;
using namespace meshlib;

CallbackHandlerC::CallbackHandlerC(
	int interval,
	const std::string & rPattern,
	const std::string & rExt,
	const std::string & rDir,
	bool verbose,
	bool silent ) throw ()
:
	m_interval(interval),
	m_pattern(rPattern),
	m_ext(rExt),
	m_dir(rDir),
	m_verbose(verbose),
	m_silent(silent),
	m_round(0),
	m_fileCounter(0),
	m_savedThisRound(false)
{
}

CallbackHandlerC::~CallbackHandlerC() throw ()
{
}

///A helper function.
template <class T>
bool isNaN( T v )
{
    return !(0.0 <= v) && !(v <= 0.0);
}

void CallbackHandlerC::handle( const DSMC::ResultsC & rResults )
{
	static const char* szNanMessage =
	"\n"
	"ERROR: Mesh energy returned NaN (Not a Number).\n"
	"This is normally caused by incorrect byte order in the input image.\n"
	"Try to add (or remove) 'image_swap_bytes' -argument in your config\n"
	"file or command line.\n";
	
	if ( m_verbose )
		cout << "Round: " << rResults.getRounds();

	if ( rResults.getOuterMesh() )
	{
		if ( isNaN( rResults.getOuterEnergy() ) )
		{
			cout << szNanMessage << endl;
			exit(-1);
		}
		
		if ( m_verbose )
			cout << " Outer V: " << rResults.getOuterMesh()->getVolumeMetric()
				<< ", E: " << rResults.getOuterEnergy();
	}

	if ( rResults.getInnerMesh() )
	{
		if ( isNaN( rResults.getInnerEnergy() ) )
		{
			cout << szNanMessage << endl;
			exit(-1);
		}
		
		if ( m_verbose )
			cout << " Inner V: " << rResults.getInnerMesh()->getVolumeMetric()
				<< ", E: " << rResults.getInnerEnergy();
	}
			
	if ( m_verbose )
		cout << endl;
	else if ( ! m_silent )
	{
		cout << "." << flush;	
	}
		
	// Save intermediate results
	m_savedThisRound = false;
	if ( m_interval > 0 && m_round == 0 )
	{
		m_round = m_interval;
		m_savedThisRound = true;
		
		saveIntermediate( rResults );		
		
		++m_fileCounter;
	}
	--m_round;		
}


/**Helper function.
 */
void CallbackHandlerC::save_mesh(
	const DSMC::ResultsC & rResults,
	const string & rFilename )
{
	MeshC * pMesh = rResults.getMesh();
	
	ostringstream os;
	os << ParamDefsC::getParams().toString();	
	os << endl << "Mesh statistics:" << endl << endl;
	os << "\tWhich: " << ((rResults.getSearchDirection() == 1)
		? "inner" : "outer") << endl;
	os << "\tMexels: " << pMesh->getMexelCount() << endl;
	os << "\tFaces: " << pMesh->getFaceCount() << endl;
	os << "\tRounds: " << rResults.getRounds() << endl;
	os << "\tEnergy: " << rResults.getEnergy() << endl;
	os << "\tVolume: " << pMesh->getVolumeMetric() << endl;
	
	MeshC::ParametersC dsmInfo( os.str() );

	MeshC imed(*pMesh);
	
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_IMAGE_COORDS) )
	{
		imed.translate( imed.getReferencePoint() );
		imed.setReferencePoint( MLVector(0.0, 0.0, 0.0) );
	}			
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_TRIANGULATED) )
		imed.triangulate();
			
	if ( ParamDefsC::getParams().getBool( ParamDefsC::OUTPUT_OFF ) )
		OFFMeshLoaderC::getInstance().save(
			&imed,
			rFilename,
			dsmInfo );
	else
		VRMLMeshLoaderC::getInstance().save(
			&imed,
			rFilename,
			dsmInfo );
}

//
void CallbackHandlerC::save_meshes(
		const dsmlib::DSMC::ResultsC & rResults,
		const std::string & rOuterFilename,
		const std::string & rInnerFilename )
{
	//save outer
	{
	MeshC * pOuter = rResults.getOuterMesh();
	
	ostringstream os;
	os << ParamDefsC::getParams().toString();	
	os << endl << "Mesh statistics:" << endl << endl;
	os << "\tWhich: outer" << endl;
	os << "\tMexels: " << pOuter->getMexelCount() << endl;
	os << "\tFaces: " << pOuter->getFaceCount() << endl;
	os << "\tRounds: " << rResults.getRounds() << endl;
	os << "\tEnergy: " << rResults.getOuterEnergy() << endl;
	os << "\tVolume: " << pOuter->getVolumeMetric() << endl;
	
	MeshC::ParametersC dsmInfo( os.str() );

	MeshC imed(*pOuter);
	
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_IMAGE_COORDS) )
	{
		imed.translate( imed.getReferencePoint() );
		imed.setReferencePoint( MLVector(0.0, 0.0, 0.0) );
	}			
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_TRIANGULATED) )
		imed.triangulate();
			
	if ( ParamDefsC::getParams().getBool( ParamDefsC::OUTPUT_OFF ) )
		OFFMeshLoaderC::getInstance().save(
			&imed,
			rOuterFilename,
			dsmInfo );
	else
		VRMLMeshLoaderC::getInstance().save(
			&imed,
			rOuterFilename,
			dsmInfo );
	}

	//save inner
	{
	MeshC * pInner = rResults.getInnerMesh();
	
	ostringstream os;
	os << ParamDefsC::getParams().toString();	
	os << endl << "Mesh statistics:" << endl << endl;
	os << "\tWhich: outer" << endl;
	os << "\tMexels: " << pInner->getMexelCount() << endl;
	os << "\tFaces: " << pInner->getFaceCount() << endl;
	os << "\tRounds: " << rResults.getRounds() << endl;
	os << "\tEnergy: " << rResults.getInnerEnergy() << endl;
	os << "\tVolume: " << pInner->getVolumeMetric() << endl;
	
	MeshC::ParametersC dsmInfo( os.str() );

	MeshC imed(*pInner);
	
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_IMAGE_COORDS) )
	{
		imed.translate( imed.getReferencePoint() );
		imed.setReferencePoint( MLVector(0.0, 0.0, 0.0) );
	}			
	if ( ParamDefsC::getParams().getBool(ParamDefsC::OUTPUT_TRIANGULATED) )
		imed.triangulate();
			
	if ( ParamDefsC::getParams().getBool( ParamDefsC::OUTPUT_OFF ) )
		OFFMeshLoaderC::getInstance().save(
			&imed,
			rInnerFilename,
			dsmInfo );
	else
		VRMLMeshLoaderC::getInstance().save(
			&imed,
			rInnerFilename,
			dsmInfo );
	}
}

void CallbackHandlerC::saveIntermediate( const DSMC::ResultsC & rResults )
{
	if ( m_interval <= 0 )
		return;
		
	// save results
	ostringstream os;
	os.width(4);
	os.fill('0');
	os << m_fileCounter;
	
	if ( rResults.getOuterMesh() && rResults.getInnerMesh() )
	{
		string outerFilename(
			m_dir + "/" + m_pattern + "_outer" + os.str() + "." + m_ext);
	
		string innerFilename(
			m_dir + "/" + m_pattern + "_inner" + os.str() + "." + m_ext);
	
		if ( m_verbose )
			cout << "Writing \'"
				<< outerFilename << "\' and \'"
				<< innerFilename << "\'..." << flush;
		
		save_meshes( rResults, outerFilename, innerFilename );
	
		if ( m_verbose )
			cout << "OK." << endl;
	}
	else
	{
		string fullFilename(
			m_dir + "/" + m_pattern + os.str() + "." + m_ext);
	
		if ( m_verbose )
			cout << "Writing \'" << fullFilename << "\'..." << flush;
		
		save_mesh( rResults, fullFilename );
	
		if ( m_verbose )
			cout << "OK." << endl;
	}
}
