/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ParamDefsC.h"

#include <iolib/ExceptionC.h>

#include <fstream>

using namespace paramlib;
using namespace std;
using namespace iolib;

ParametersC * ParamDefsC::m_pParams = NULL;

void ParamDefsC::initParams( int argc, char * argv[] )
{
	// parameters initialized to their default values
	ParametersC defaultParams( ParamDefsC::getDefs() );

	// if no arguments, set help=true
	if ( argc == 1 )
	{
		argc = 2;
		static char * argv2[] = {"", "--help"};
		argv = argv2;
	}

	// parameters from the command line
	ParametersC cmdLineParams( ParamDefsC::getDefs() , argc, argv );

	// command-line parameters override the defaults
	ParametersC params = defaultParams + cmdLineParams;

	// use config file if present
	if ( params.containsKey( ParamDefsC::CONFIG ) )
	{
		ifstream configFile(
			params.getString( ParamDefsC::CONFIG ).c_str() );
			
		if ( configFile.good() )
		{
			// config file overrides the defaults,
			// command line overrides both.
			params = 
				defaultParams + 
				ParametersC ( ParamDefsC::getDefs(), configFile ) +
				cmdLineParams;
		}
		configFile.close();		
	}
		
	m_pParams = new ParametersC( params );
}

paramlib::ParametersC & ParamDefsC::getParams()
{
	assert( m_pParams );
	return *m_pParams;
}

// parameter definition format is:
//{
//	ID,
//	type,
//	names,
//	aliases,
//	valuedescription,
//	description,
//	defaultvalue,
//	overrides,
//}
//

/**A static initializer for parameter definitions
 */
ParametersC::DefinitionS ParamDefsC::definitions[] =
{
	{
		OUTER_MESH,
		ParametersC::STRING,
		"outer_mesh",
		"M",
		"<filename>",
		"Use initial outer mesh from <filename>",
		NULL,
		{ INNER_MESH, OUTER_SCALE, OUTER_REF },
	},
	{
		OUTER_SCALE,
		ParametersC::VECTOR,
		"outer_scale",
		"S",
		"<x> <y> <z>",
		"Scale initial outer mesh",
		"1.0 1.0 1.0",
		{},
	},
	{
		OUTER_REF,
		ParametersC::VECTOR,
		"outer_ref",
		"R",
		"<x> <y> <z>",
		"Set reference point for outer mesh",
		NULL,
		{},
	},
	{
		INNER_MESH,
		ParametersC::STRING,
		"inner_mesh",
		"m",
		"<filename>",
		"Use initial inner mesh from <filename>",
		NULL,
		{ OUTER_MESH, INNER_SCALE, INNER_REF },
	},
	{
		INNER_SCALE,
		ParametersC::VECTOR,
		"inner_scale",
		"s",
		"<x> <y> <z>",
		"Scale initial inner mesh",
		"1.0 1.0 1.0",
		{},
	},
	{
		INNER_REF,
		ParametersC::VECTOR,
		"inner_ref",
		"r",
		"<x> <y> <z>",
		"Set reference point for inner mesh",
		NULL,
		{},
	},
	{
		IMAGE,
		ParametersC::STRING,
		"image",
		"i",
		"<filename>",
		"Use energy image from <filename>",
		NULL,
		{IMAGE_DIMENSIONS, IMAGE_DATA},
	},
	{
		IMAGE_DIMENSIONS,
		ParametersC::VECTOR,
		"image_dimensions",
		"d",
		"<x> <y> <z>",
		"Specify input image dimensions for raw voxel data",
		NULL,
		{IMAGE, IMAGE_DATA},
	},
	{
		IMAGE_DATA,
		ParametersC::STRING,
		"image_data",
		"a",
		"float | int | uint",
		"specify input image data type for raw voxel data",
		NULL,
		{IMAGE, IMAGE_DIMENSIONS},
	},
	{
		IMAGE_SWAP_BYTES,
		ParametersC::BOOL,
		"image_swap_bytes",
		"b",
		"true | false",
		"Swap input image byte order",
		"false",
		{},
	},
	{
		IMAGE_NORMALIZE,
		ParametersC::BOOL,
		"image_normalize",
		"z",
		"true | false",
		"Normalize image data",
		"true",
		{},
	},
	{
		OUTPUT_MESH,
		ParametersC::STRING,
		"output_mesh",
		"o",
		"<filename>",
		"Store output mesh in <filename>",
		"result.wrl",
		{},
	},
	{
		OUTPUT_OFF,
		ParametersC::BOOL,
		"output_off",
		"",
		"true | false",
		"Store output mesh in OFF format",
		"false",
		{},
	},
	{
		OUTPUT_IMAGE_COORDS,
		ParametersC::BOOL,
		"output_image_coords",
		"",
		"true | false",
		"Store output mesh in image coordinates",
		"false",
		{},
	},
	{
		OUTPUT_TRIANGULATED,
		ParametersC::BOOL,
		"output_triangulated",
		"",
		"true | false",
		"Triangulate output mesh",
		"false",
		{},
	},
	{
		IMED_INTERVAL,
		ParametersC::INT,
		"imed_interval",
		"y",
		"#",
		"Store intermediate results every # steps",
		"-1",
		{},
	},
	{
		IMED_PATTERN,
		ParametersC::STRING,
		"imed_pattern",
		"w",
		"<pattern>",
		"Use <pattern> to name intermediate result files",
		"result",
		{},
	},
	{
		IMED_DIR,
		ParametersC::STRING,
		"imed_dir",
		"x",
		"<dir>",
		"Store intermediate results in directory <dir>",
		"./",
		{},
	},
	{
		MAX_ROUNDS,
		ParametersC::INT,
		"max_rounds",
		"N",
		"#",
		"Set maximum rounds to #",
		"50",
		{},
	},
	{
		TOLERANCE,
		ParametersC::FLOAT,
		"tolerance",
		"t",
		"<value>",
		"Set tolerance to <value>",
		"1.05",
		{},
	},
	{
		VOLUME_RATIO,
		ParametersC::FLOAT,
		"volume_ratio",
		"",
		"<value>",
		"Set target volume ratio for single-surface method",
		"0.3",
		{},
	},
	{
		LAMBDA,
		ParametersC::FLOAT,
		"lambda",
		"l",
		"<value>",
		"Set regularization parameter to <value>",
		"0.3",
		{},
	},
	{
		GAMMA,
		ParametersC::FLOAT,
		"gamma",
		"g",
		"<value>",
		"Set penalty coefficient to <value>",
		"0.03",
		{},
	},
	{
		SEARCH_J,
		ParametersC::INT,
		"search_j",
		"J",
		"#",
		"Set search space parameter J to #",
		"8",
		{},
	},
	{
		SEARCH_K,
		ParametersC::INT,
		"search_k",
		"K",
		"#",
		"Set search space parameter K to #",
		"1",
		{},
	},
	{
		SEARCH_L,
		ParametersC::FLOAT,
		"search_l",
		"L",
		"<value>",
		"Set search space parameter L to <value>",
		"0.5",
		{},
	},
	{
		SEARCH_D,
		ParametersC::FLOAT,
		"search_d",
		"D",
		"<value>",
		"Set search space parameter L to <value>",
		"0.15",
		{},
	},
	{
		HELP,
		ParametersC::NONE,
		"help usage",
		"h?",
		"",
		"Output help screen",
		NULL,
		{VERBOSE},
	},
	{
		VERBOSE,
		ParametersC::BOOL,
		"verbose",
		"v",
		"true | false",
		"Enable/disable verbose mode",
		"false",
		{},
	},
	{
		SILENT,
		ParametersC::BOOL,
		"silent quiet",
		"q",
		"true | false",
		"Enable/disable silent mode",
		"false",
		{},
	},
	{
		CONFIG,
		ParametersC::STRING,
		"config",
		"c",
		"<filename>",
		"Use configuration file <filename>",
		"config.dsm",
		{},
	}
};

std::vector<paramlib::ParametersC::DefinitionS> ParamDefsC::getDefs()
{
	return std::vector<paramlib::ParametersC::DefinitionS> ( definitions,
		definitions + sizeof( definitions ) / sizeof( definitions[0] ) );
};
