/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//
// documentation.h
// 
// This file contains doxygen documentation pages.
//
//

/** @mainpage Dual Surface Minimization in C++
 * 
 * @section intro_sec Introduction
 * 
 * Dual Surface Minimization (DSM) @ref ref_1 "[1]" is a
 * fully automatic method for extracting meaningful structures from
 * three-dimensional image data.
 * @a dualsurfacemin is a C++ implementation of the DSM
 * algorithm. This chapter describes the usage of the
 * @a dualsurfacemin tool in detail.
 * 
 * @section usage_sec Usage
 * 
 * Parameters can be passed on the command line or via a configuration file
 * ('config.dsm').
 * When the program is launched, parameters will be initialized to their
 * hard-coded default values. If a valid configuration file is present, the values
 * specified in the file will override the defaults. Arguments specified on the
 * command-line override the default values and values specified in the
 * configuration file.
 * 
 * @subsection cmdline_sec Simple command-line usage
 * 
 * To get command-line help launch one of the following:\n
 * \n
 * <tt>dualsurfacemin --help</tt>\n
 * <tt>dualsurfacemin --usage</tt>\n
 * <tt>dualsurfacemin -h</tt>\n
 * <tt>dualsurfacemin -?</tt>\n
 * \n
 * To get a more detailed help screen, use the --verbose argument:\n
 * \n
 * <tt>dualsurfacemin --help --verbose</tt>\n
 * \n
 * To perform outer surface minimization with initial outer mesh
 * 'outer.wrl', image 'hoffmann' and store results as
 * 'hoffsurf.wrl', launch:\n
 * \n
 *  <tt>dualsurfacemin --outer_mesh outer.wrl --image hoffmann --output_mesh hoffsurf.wrl</tt>\n
 * \n
 * The most commonly used parameters have short command-line aliases. To
 * achieve the same results as in the example above, one could launch:\n
 * \n
 *  <tt>dualsurfacemin -M outer.wrl -i hoffmann -o hoffsurf.wrl</tt>\n
 * \n
 * Note that when only one input mesh is defined (as in the example above)
 * @a dualsurfacemin will use a single surface version of the algorithm
 * (DSM-OS or DSM-IS). When both input meshes are defined, the standard dual
 * surface algorithm will be used.
 * 
 * See @ref params_sec "parameter summary" for supported command-line
 * arguments and their aliases.
 * 
 * @subsection input_sec Input files
 * 
 * @subsubsection meshfiles_sec Mesh files
 * 
 * Currently supported mesh file formats are VRML 1.0 @ref ref_2 "[2]" and OFF @ref ref_3 "[3]".
 * There are certain constraints for the meshes:
 * - Input meshes must be simplex (each vertex must be connected to exactly
 * three neighbouring vertices by an edge);
 * - Faces must be properly oriented (vertex indices must be listed
 * couter-clockwise);
 * - Only a certain subset of VRML 1.0 is supported, see sample VRML files
 * for details.
 * 
 * Input meshes are assumed to be in surface-centered coordinates.
 * If the input mesh is in OFF format, the reference point in image
 * coordinates must be supplied with @c --outer_ref or @c --inner_ref.
 * If the input mesh is in VRML, the reference point will be read from an
 * enclosing @a Translation node in the VRML file.
 * 
 * @subsubsection imgfiles_sec Image files
 * 
 * @paragraph analyze_par Analyze 7.5
 * 
 * If image filename is supplied without extra arguments, the file is assumed
 * to be in Analyze 7.5 format @ref ref_4 "[4]". The @a filename argument to the @c --image
 * parameter should be the Analyze file name without extension.
 * 
 * @paragraph raw_par Raw voxel data
 * 
 * If the input image is raw voxel data, @c --image_dimensions and
 * @c --image_data parameters must be supplied.
 * 
 * @subsection output_sec Program output
 * 
 * @subsubsection omesh_sec Output mesh
 * 
 * By default the output mesh will be in VRML 1.0 format, simplex, and in
 * surface-centered coordinates. Output format can be changed to .OFF with
 * @c --output_off. Triangulated output can be produced with
 * @c --output_triangulated. Output can be transformed to image coordinates
 * with @c --output_image_coords.
 * 
 * @subsubsection imed_sec Intermediate results
 * 
 * @a dualsurfacemin can produce intermediate result meshes while the
 * algorithm is running. For example to produce intermediate meshes
 * every 5 rounds of the algorithm and have these meshes stored in
 * subdirectory <code>./Intermed</code>, one could launch the program with the
 * following arguments:\n
 * \n
 * <code> --imed_interval 5 --imed_dir ./Intermed --imed_pattern foobar</code>\n
 * \n
 * This would produce files @c foobar0001.wrl, @c foobar0002.wrl etc. in
 * directory <code>./Intermed</code>. Note that @c --imed_dir and
 * @c --imed_pattern are not mandatory;
 * If these arguments are not supplied, current directory and pattern
 * @c result will be used. Note also that the directory defined with
 * @c --imed_dir must exist; @a dualsurfacemin will not attempt to create one
 * automatically.
 * 
 * @subsubsection co_sec Console output
 * 
 * By default some basic console output will be produced (such as 'input image
 * loaded ok blah blah...'). If the program is invoked with the @c --verbose
 * parameter, detailed minimization statistics will be produced every round of
 * the algorithm. If the program is invoked with the @c --silent parameter, all
 * console output is suppressed.
 * 
 * @subsection configfile_sec Configuration files
 * 
 * When @a dualsurfacemin is launched, it will look for a file called
 * 'config.dsm' (unless @c --config command-line argument was used; see
 * @ref params_sec "parameter summary").
 * A configuration file is a plain text file which contains key/value pairs
 * separated by newline characters. Comments are allowed and
 * are prefixed by hash (@c #) characters. Below is an example of a valid
 * configuration file:\n
 * \n
 * <code>
 * # My DSM config file\n
 * # input:\n
 * inner_mesh = SmallBall.wrl\n
 * outer_mesh = BigBall.wrl\n
 * image = energy.img\n
 * image_dimensions = 128 128 35\n
 * image_normalize = true\n
 * image_swap_bytes = true\n
 * image_data = float\n
 * \n
 * lambda = 0.4\n
 * max_rounds = 1500\n
 *\n
 * # output:\n
 * output_mesh = BrainHertz.wrl\n
 * </code>
 * \n
 * Configuration file keys are generally syntactically identical to the
 * corresponding command-line arguments, but there are some arguments that
 * are not meaningful in the config file context (such as @c --help).
 * Note also that short command-line aliases are not valid config file keys.
 * See @ref params_sec "parameter summary" for a table of supported config file
 * keys.\n
 * \n
 * <b>Note:</b> As command-line arguments override config file entries,
 * specifying a single input mesh on the command line will always result in the
 * single surface algorithm being used, even if input meshes are defined in the
 * config file. For example if the config file contains an entry for
 * @c inner_mesh, and the user specifies the @c --outer_mesh parameter on
 * command-line, @a dualsurfacemin will use the single surface algorithm
 * DSM-OS instead of the standard dual surface algorithm.
 * 
 * @subsection params_sec Parameter summary
 *
 * Parameters are passed as key/value pairs. Command-line usage is:\n
 * \n
 * <code>dualsurfacemin --<i>key_name</i> [<i>value</i>]</code> or\n 
 * <code>dualsurfacemin -<i>a</i> [<i>value</i>]</code>,
 * where <i>a</i> is a short command-line alias for the key.\n
 * \n
 * Config file usage is:\n
 * \n
 * <code><i>key_name</i> = <i>value</i></code>\n
 * \n
 * On the command-line, for keys which take boolean values, specifying
 * <code>--key_name</code> is equivalent to <code>--key_name true</code>.
 * In the configuration file the boolean value (@c true or @c false) must be
 * written out explicitly.\n
 * \n
 * All currently supported parameter keys and their aliases are described in
 * the following tables.\n
 * \n
 * <small>
 * <table>
 * <tr>
 *  <td><b>Category</b>
 *  <td><b>Key</b>
 *  <td><b>Alias</b>
 *  <td><b>Value</b>
 *  <td><b>Description</b>
 *  <td><b>Mandatory</b>
 *  <td><b>Default value</b>
 * 
 * <tr>
 *  <td><b>Input</b>
 *  <td>outer_mesh
 *  <td>M
 *  <td><i>filename</i>
 *  <td>Use initial outer mesh from file <i>filename</i>
 *  <td>yes<sup>@ref ps_ref_a "a"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>outer_scale
 *  <td>S
 *  <td><i>ratio</i>
 *  <td>Scale initial outer mesh by <i>ratio</i>
 *  <td>no
 *  <td>1.0
 * 
 * <tr>
 *  <td>
 *  <td>outer_ref
 *  <td>R
 *  <td><i>x y z</i>
 *  <td>Set reference point for the initial outer mesh
 *  <td><sup>@ref ps_ref_b "b"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>inner_mesh
 *  <td>m
 *  <td><i>filename</i>
 *  <td>Use initial inner mesh from file <i>filename</i>
 *  <td>yes<sup>@ref ps_ref_a "a"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>inner_scale
 *  <td>s
 *  <td><i>ratio</i>
 *  <td>Scale initial inner mesh by <i>ratio</i>
 *  <td>no
 *  <td>1.0
 * 
 * <tr>
 *  <td>
 *  <td>inner_ref
 *  <td>r
 *  <td><i>x y z</i>
 *  <td>Set reference point for the initial inner mesh
 *  <td><sup>@ref ps_ref_b "b"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>image
 *  <td>i
 *  <td><i>filename</i>
 *  <td>Use energy image from file <i>filename</i>
 *  <td>yes
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>image_dimensions
 *  <td>d
 *  <td><i>x y z</i>
 *  <td>Set input image resolution
 *  <td><sup>@ref ps_ref_c "c"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>image_data
 *  <td>a
 *  <td>float | int | uint
 *  <td>Set input image data type to floating point, integer or unsigned integer
 *  <td><sup>@ref ps_ref_c "c"</sup>
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>image_swap_bytes
 *  <td>b
 *  <td>true | false
 *  <td>Swap input image byte order
 *  <td>no
 *  <td>false
 * 
 * <tr>
 *  <td>
 *  <td>image_normalize
 *  <td>z
 *  <td>true | false
 *  <td>Normalize input image data
 *  <td>no
 *  <td>true
 * 
 * </table>
 * @anchor ps_ref_a <sup>a</sup>At least one input mesh must be supplied.\n
 * @anchor ps_ref_b <sup>b</sup>A reference point must be supplied if it is not
 * present in the input VRML file.\n
 * @anchor ps_ref_c <sup>c</sup>Mandatory for raw images. When this parameter
 * is specified the input image will be treated as raw.\n
 * \n
 * 
 * <table>
 * <tr>
 *  <td><b>Category</b>
 *  <td><b>Key</b>
 *  <td><b>Alias</b>
 *  <td><b>Value</b>
 *  <td><b>Description</b>
 *  <td><b>Mandatory</b>
 *  <td><b>Default value</b>
 * 
 * <tr>
 *  <td><b>Output</b>
 *  <td>output_mesh
 *  <td>o
 *  <td><i>filename</i>
 *  <td>Store resulting mesh in file <i>filename</i>
 *  <td>no
 *  <td>result.wrl
 * 
 * <tr>
 *  <td>
 *  <td>output_off
 *  <td>
 *  <td>true | false
 *  <td>Store resulting mesh in OFF format
 *  <td>no
 *  <td>false
 * 
 * <tr>
 *  <td>
 *  <td>output_triangulated
 *  <td>
 *  <td>true | false
 *  <td>Triangulate the resulting mesh
 *  <td>no
 *  <td>false
 * 
 * <tr>
 *  <td>
 *  <td>output_image_coords
 *  <td>
 *  <td>true | false
 *  <td>Translate the resulting mesh to image coordinates
 *  <td>no
 *  <td>false
 * 
 * <tr>
 *  <td>
 *  <td>imed_interval
 *  <td>y
 *  <td><i>#</i>
 *  <td>Store intermediate results every <i>#</i> rounds
 *  <td>no
 *  <td>0 (Intermediate results are not stored by default)
 * 
 * <tr>
 *  <td>
 *  <td>imed_pattern
 *  <td>w
 *  <td><i>pattern</i>
 *  <td>Use <i>pattern</i> to name intermediate result files. For example if imed_pattern is set to 'foo', intermediate result files will be named 'foo0001.wrl', 'foo0002.wrl' etc.
 *  <td>no
 *  <td>'result'
 * 
 * <tr>
 *  <td>
 *  <td>imed_dir
 *  <td>x
 *  <td><i>dirname</i>
 *  <td>Store intermediate results to directory <i>dirname</i>. If the specified directory does not exist, the program will abort
 *  <td>no
 *  <td>./
 * 
 * </table>
 * \n\n
 * <table>
 * <tr>
 *  <td><b>Category</b>
 *  <td><b>Key</b>
 *  <td><b>Alias</b>
 *  <td><b>Value</b>
 *  <td><b>Description</b>
 *  <td><b>Mandatory</b>
 *  <td><b>Default value</b>
 * 
 * <tr>
 *  <td><b>Minimization</b>
 *  <td>max_rounds
 *  <td>N
 *  <td><i>#</i>
 *  <td>Set maximum rounds to <i>#</i>
 *  <td>no
 *  <td>50
 * 
 * <tr>
 *  <td>
 *  <td>tolerance
 *  <td>t
 *  <td><i>value</i>
 *  <td>Set tolerance (in voxels) when comparing mesh volumes<sup>@ref ps_ref_d "d"</sup>
 *  <td>no
 *  <td>0.5
 * 
 * <tr>
 *  <td>
 *  <td>volume_ratio
 *  <td>
 *  <td><i>value</i>
 *  <td>Set target volume ratio for single surface method<sup>@ref ps_ref_e "e"</sup>
 *  <td>no
 *  <td>0.6
 * 
 * <tr>
 *  <td>
 *  <td>lambda
 *  <td>l
 *  <td><i>value</i>
 *  <td>Set regularization parameter to <i>value</i>
 *  <td>no
 *  <td>0.3
 * 
 * <tr>
 *  <td>
 *  <td>gamma
 *  <td>g
 *  <td><i>value</i>
 *  <td>Set penalty coefficient to <i>value</i>
 *  <td>no
 *  <td>0.03
 * 
 * <tr>
 *  <td>
 *  <td>search_j
 *  <td>J
 *  <td><i>#</i>
 *  <td>Set search space parameter J to <i>#</i>
 *  <td>no
 *  <td>8
 * 
 * <tr>
 *  <td>
 *  <td>search_k
 *  <td>K
 *  <td><i>#</i>
 *  <td>Set search space parameter K to <i>#</i>
 *  <td>no
 *  <td>1
 * 
 * <tr>
 *  <td>
 *  <td>search_l
 *  <td>L
 *  <td><i>value</i>
 *  <td>Set search space parameter L to <i>value</i>
 *  <td>no
 *  <td>0.5
 * 
 * <tr>
 *  <td>
 *  <td>search_d
 *  <td>D
 *  <td><i>value</i>
 *  <td>Set search space parameter D to <i>value</i>
 *  <td>no
 *  <td>0.15
 * 
 * </table>
 * @anchor ps_ref_d <sup>d</sup>This parameter is meaningful only for the dual surface method.\n
 * @anchor ps_ref_e <sup>e</sup>This parameter is meaningful only for the single surface method.\n
 * \n
 * 
 * <table>
 * <tr>
 *  <td><b>Category</b>
 *  <td><b>Key</b>
 *  <td><b>Alias</b>
 *  <td><b>Value</b>
 *  <td><b>Description</b>
 *  <td><b>Mandatory</b>
 *  <td><b>Default value</b>
 * 
 * <tr>
 *  <td><b>Console output</b>
 *  <td>help, usage
 *  <td>h, ?
 *  <td>
 *  <td>Output a help screen to the console. This key is not allowed in the configuration file.
 *  <td>no
 *  <td>
 * 
 * <tr>
 *  <td>
 *  <td>verbose
 *  <td>v
 *  <td>true | false
 *  <td>Enable/disable verbose mode. Extra console output is produced in verbose mode.
 *  <td>no
 *  <td>false
 * 
 * <tr>
 *  <td>
 *  <td>silent, quiet
 *  <td>q
 *  <td>true | false
 *  <td>Enable/disable silent mode. No console output whatsoever is produced in silent mode.
 *  <td>no
 *  <td>false
 * 
 * </table>
 * \n\n
 * <table>
 * <tr>
 *  <td><b>Category</b>
 *  <td><b>Key</b>
 *  <td><b>Alias</b>
 *  <td><b>Value</b>
 *  <td><b>Description</b>
 *  <td><b>Mandatory</b>
 *  <td><b>Default value</b>
 * 
 * <tr>
 *  <td><b>Miscellaneous</b>
 *  <td>config
 *  <td>c
 *  <td><i>filename</i>
 *  <td>Use configuration file <i>filename</i>
 *  <td>no
 *  <td>config.dsm
 * 
 * </table>
 * </small>
 * 
 * 
 * @section refs_sec References
 * 
 * -# @anchor ref_1 J. Tohka, J. Mykk&auml;nen, "Deformable mesh for automated surface extraction from noisy images", 
 * International Journal of Image and Graphics, Vol. 4, No.3 (2004) 405-432
 * -# @anchor ref_2 Web 3D consortium <http://www.web3d.org/>
 * -# @anchor ref_3 Geomview <http://www.geomview.org/>
 * -# @anchor ref_4 Mayo Biomedical Imaging Resource <http://www.mayo.edu/bir/>
 * 
 */

