/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "documentation.h"

/**@file main.cpp
 * The main program.
 */
 
#include "ParamDefsC.h"
#include "CallbackHandlerC.h"

#include <dsmlib/dsmlib.h>
#include <imagelib/imagelib.h>
#include <meshlib/meshlib.h>
#include <paramlib/paramlib.h>
#include <iolib/iolib.h>

#include <meshlib/VRMLMeshLoaderC.h> ///@todo: remove this
#include <meshlib/OFFMeshLoaderC.h> ///@todo: remove this
#include <imagelib/RAWImageLoaderC.h> ///@todo: remove this
#include <imagelib/AnalyzeImageLoaderC.h> ///@todo: remove this

#include <iostream>
#include <fstream>
#include <sstream>

using namespace dsmlib;
using namespace imagelib;
using namespace meshlib;
using namespace paramlib;
using namespace iolib;
using namespace std;

/**A helper function.
 * Outputs a string to the console only if in verbose mode.
 * Do not call before initParams!
 */
void coutVerbose( const string & rString )
{
	if ( ParamDefsC::getParams().getBool( ParamDefsC::VERBOSE ) )
		cout << rString << flush;
}

/**A helper function.
 * Outputs a string to the console only if not in silent mode.
 * Do not call before initParams!
 */
void coutNormal( const string & rString )
{
	if ( ! ParamDefsC::getParams().getBool( ParamDefsC::SILENT ) )
		cout << rString << flush;
}

/**A helper function.
 * Outputs a detailed help screen to the console.
 * The help screen is constructed from parameter definitions.
 * Do not call before initParams!
 */
void outputDetailedHelpScreen()
{
/*	INCOMPLETE
   std::vector<ParametersC::DefinitionS> defs = ParamDefsC::getParamDefs();
	std::vector<ParametersC::DefinitionS>::const_iterator itDef =
		defs.begin();
	while ( itDef != defs.end() )
	{
		
		++itDef;
	}*/
	
	cout <<
"Usage:\n"
"  dualsurfacemin [ option [arguments...] ...\n"
"\n"
"Valid options include:\n"
"\n"
"Option           Alias  Arguments      Description                  Defaults to\n"
"-------------------------------------------------------------------------------\n"
"--config            -c  <filename>     Use config file             'config.dsm'\n"
"\n"
"--outer_mesh        -M  <filename>     Outer input mesh\n"
"--outer_scale       -S  <x y z>        Outer mesh scale factor            1 1 1\n"
"--outer_ref         -R  <x y z>        Outer mesh reference point\n"
"--inner_mesh        -m  <filename>     Inner input mesh\n"
"--inner_scale       -s  <x y z>        Inner mesh scale factor            1 1 1\n"
"--inner_ref         -r  <x y z>        Inner mesh reference point\n"
"--image             -i  <filename>     Input image\n"
"--image_dimensions  -d  <x y z>        Image dimensions\n"
"--image_data        -a  float|int|uint Image data type\n"
"--image_swap_bytes  -b  true|false     Swap image endianess               false\n"
"--image_normalize   -z  true|false     Normalize image                     true\n"
"\n"
"--output_mesh       -o  <filename>     Output mesh\n"
"--output_off            true|false     Write .OFF format                  false\n"
"--output_triangulated   true|false     Write triangulated mesh            false\n"
"--output_image_coords   true|false     Write mesh in image coords         false\n"
"--imed_interval     -y  <#>            Store results every # rounds          -1\n"
"--imed_pattern      -w  <pattern>      Pattern for imed filess         'result'\n"
"--imed_dir          -x  <dirname>      Directory for imed results            ./\n"
"\n"
"--max_rounds        -N  <#>            Terminate after # rounds              50\n"
"--tolerance         -t  <value>        Volume tolerance for DSM            1.05\n"
"--volume_ratio          <value>        Volume ratio for DSM-OS              0.6\n"
"--lambda            -l  <value>        Regularization parameter             0.3\n"
"--gamma             -g  <value>        Penalty coefficient                 0.03\n"
"--search_j          -J  <#>            Search space parameter J               8\n"
"--search_k          -K  <#>            Search space parameter K               1\n"
"--search_l          -L  <value>        Search space parameter L             0.5\n"
"--search_d          -D  <value>        Search space parameter D            0.15\n"
"\n"
"--help, --usage     -h,-?              Console help screen\n"
"--verbose           -v                 Enable verbose output              false\n"
"--silent, --quiet   -q                 Enable quiet mode                  false\n"
"\n";
	
}

/**The main function.
 *@todo This is getting pretty big. Divide!
 */
int main(int argc, char * argv[] )
{
	
	/// @todo Are file name extensions mandatory?
	/// @todo Detailed help screen.
		
	// help screen strings
	static const char * szUsage =
"Usage:\n"
"  dualsurfacemin --outer_mesh <file> --inner_mesh <file> --image <file>\n"
"\n"
"For a more detailed help screen launch:\n"
"  dualsurfacemin --help --verbose\n"
"\n";

	try {
	try {
	try {
		
		// mesh & image pointers.
		ImageI * pImage = NULL;
		MeshC* pOuterMesh = NULL;
		MeshC* pInnerMesh = NULL;
		
				
/***************************************************************************
 * 
 * Initialize parameter set
 * 
 ***************************************************************************/

		ParamDefsC::initParams( argc, argv );

		coutVerbose("\ndualsurfacemin v1.1.1\n\n");
				
/***************************************************************************
 * 
 * Print help screen if needed
 * 
 ***************************************************************************/

		if ( ParamDefsC::getParams().containsKey( ParamDefsC::HELP ) )
		{
			if ( ParamDefsC::getParams().containsKey( ParamDefsC::VERBOSE ) )
				outputDetailedHelpScreen();
			else
				cout << szUsage << endl;
			return 0;
		}

/***************************************************************************
 * 
 * Check for mandatory parameters
 * 
 ***************************************************************************/
	
	if ( ! ParamDefsC::getParams().containsKey( ParamDefsC::IMAGE ) )
		throw ExceptionC("Input image must be specified.");
	
	if ( ! ( ParamDefsC::getParams().containsKey( ParamDefsC::OUTER_MESH ) || 
			 ParamDefsC::getParams().containsKey( ParamDefsC::INNER_MESH ) ) )
		throw ExceptionC("At least one input mesh must be specified.");
		
	if ( ParamDefsC::getParams().containsKey(ParamDefsC::IMAGE_DATA) !=
		 ParamDefsC::getParams().containsKey(ParamDefsC::IMAGE_DIMENSIONS) )
		 throw ExceptionC( "Both \'image_data\' and \'image_dimensions\' must be specified for raw voxel images." );
		
/***************************************************************************
 * 
 * Initialize mesh and image factories
 * 
 ***************************************************************************/
	
		coutVerbose("Initializing image loaders...");

		/// @todo: Loader registration should be automated.
		ImageFactoryC & rImageFactory = ImageFactoryC::getInstance();
		rImageFactory.addLoader( RAWImageLoaderC::getInstance() );
		rImageFactory.addLoader( AnalyzeImageLoaderC::getInstance() );

		coutVerbose("OK.\n");
				
		coutVerbose("Initializing mesh loaders...");
		
		MeshFactoryC & rMeshFactory = MeshFactoryC::getInstance();
		rMeshFactory.addLoader( OFFMeshLoaderC::getInstance() );
		rMeshFactory.addLoader( VRMLMeshLoaderC::getInstance() );		

		coutVerbose("OK.\n");
						
/***************************************************************************
 * 
 * Load input image
 * 
 ***************************************************************************/

		coutNormal("Loading input image \'");
		coutNormal(ParamDefsC::getParams().getString( ParamDefsC::IMAGE ));
		coutNormal("\'...");

		// check whether we use raw voxel input
		if ( ParamDefsC::getParams().containsKey(
			ParamDefsC::IMAGE_DIMENSIONS ) )
		{
			coutVerbose("\n\tUsing raw voxel input.\n");
			// check image_data argument
			ImageI::DataTypeE imageDataType;

			const string  & rImageDataTypeString(
				ParamDefsC::getParams().getString( ParamDefsC::IMAGE_DATA ) );
				
			if ( rImageDataTypeString == "float" )
				imageDataType = ImageI::FLOAT;
			else if ( rImageDataTypeString == "int" )
				imageDataType = ImageI::INT;
			else if ( rImageDataTypeString == "uint" )
				imageDataType = ImageI::UINT;
			else
				throw ExceptionC("Unrecognized \'image_data\' argument.");
			
			// initialize image loader parameters
			ImageI::ParametersC imageParams(
				ParamDefsC::getParams().getVector(
					ParamDefsC::IMAGE_DIMENSIONS ),
				imageDataType,
				ParamDefsC::getParams().getBool(
					ParamDefsC::IMAGE_SWAP_BYTES ) );
			
			// load image
			pImage = rImageFactory.create(
				ParamDefsC::getParams().getString( ParamDefsC::IMAGE ),
				imageParams );			
		}
		else
		{
		 	// Input image is in Analyze format.
		 	
		 	coutVerbose("\n\tUsing Analyze (tm) 7.5 format.\n");

			// initialize image loader parameters
			//dimensions & data type are ignored by the Analyze loader
			ImageI::ParametersC imageParams(
				MLVector(0,0,0),
				ImageI::FLOAT,
				ParamDefsC::getParams().getBool(
					ParamDefsC::IMAGE_SWAP_BYTES ) );

			// load image
			pImage = rImageFactory.create(
				ParamDefsC::getParams().getString( ParamDefsC::IMAGE ),
				imageParams );	
		}
		if ( pImage == NULL)
			throw ExceptionC("Error while loading input image!");
		coutNormal("OK.\n");
		
		// normalize image if needed
		coutVerbose("Normalizing image...");
		if ( ParamDefsC::getParams().getBool( ParamDefsC::IMAGE_NORMALIZE ) )
			pImage->normalize();
		coutVerbose("OK.\n");

/***************************************************************************
 * 
 * Load input mesh(es)
 * 
 ***************************************************************************/

		if ( ParamDefsC::getParams().containsKey( ParamDefsC::OUTER_MESH ) )
		{
			coutNormal("Loading input mesh \'");
			coutNormal(
				ParamDefsC::getParams().getString( ParamDefsC::OUTER_MESH ));
			coutNormal("\'...");
			pOuterMesh = rMeshFactory.create(
				ParamDefsC::getParams().getString( ParamDefsC::OUTER_MESH ));
			if ( pOuterMesh == NULL )
				throw ExceptionC("Error loading input mesh!");
			coutNormal("OK.\n");
			
			if ( ! pOuterMesh->isSimplex() )
				throw ExceptionC ("Outer mesh is not simplex!");
			
			if ( ParamDefsC::getParams().containsKey( ParamDefsC::OUTER_REF ) )
				pOuterMesh->setReferencePoint(
					ParamDefsC::getParams().getVector( ParamDefsC::OUTER_REF )
					);

			if ( ParamDefsC::getParams().containsKey( ParamDefsC::OUTER_SCALE ) )
				pOuterMesh->scale(
					ParamDefsC::getParams().getVector( ParamDefsC::OUTER_SCALE )
					);
		}

		if ( ParamDefsC::getParams().containsKey( ParamDefsC::INNER_MESH ) )
		{
			coutNormal("Loading input mesh \'");
			coutNormal(
				ParamDefsC::getParams().getString( ParamDefsC::INNER_MESH ));
			coutNormal("\'...");
			pInnerMesh = rMeshFactory.create(
				ParamDefsC::getParams().getString( ParamDefsC::INNER_MESH ));
			if ( pInnerMesh == NULL )
				throw ExceptionC("Error loading input mesh!");
			coutNormal("OK.\n");

			if ( ! pInnerMesh->isSimplex() )
				throw ExceptionC ("Inner mesh is not simplex!");

			if ( ParamDefsC::getParams().containsKey( ParamDefsC::INNER_REF ) )
				pInnerMesh->setReferencePoint(
					ParamDefsC::getParams().getVector( ParamDefsC::INNER_REF )
					);

			if ( ParamDefsC::getParams().containsKey( ParamDefsC::INNER_SCALE ) )
				pInnerMesh->scale(
					ParamDefsC::getParams().getVector( ParamDefsC::INNER_SCALE )
					);
		}

		
/***************************************************************************
 * 
 * Perform energy minimization
 * 
 ***************************************************************************/
		// construct minimizer object
		coutVerbose("Initializing minimizer...");
		DSMC dsm( pOuterMesh, pInnerMesh, pImage );
		
		// register callback;
		/// @todo Check imed_* parameter validity.
		///@todo Save initial and final meshes.
		CallbackHandlerC myCallback(
			ParamDefsC::getParams().getInt( ParamDefsC::IMED_INTERVAL ),
			ParamDefsC::getParams().getString( ParamDefsC::IMED_PATTERN ),
			ParamDefsC::getParams().getBool( ParamDefsC::OUTPUT_OFF )
				? "off" : "wrl",
			ParamDefsC::getParams().getString( ParamDefsC::IMED_DIR ),
			ParamDefsC::getParams().getBool( ParamDefsC::VERBOSE ),
			ParamDefsC::getParams().getBool( ParamDefsC::SILENT ));

		dsm.setCallback( &myCallback );
		
		// initialize algorithm parameters
		DSMC::ParametersC dsmParams(
			ParamDefsC::getParams().getFloat( ParamDefsC::LAMBDA ),
			ParamDefsC::getParams().getInt( ParamDefsC::SEARCH_J ),
			ParamDefsC::getParams().getInt( ParamDefsC::SEARCH_K ),
			ParamDefsC::getParams().getFloat( ParamDefsC::SEARCH_L ),
			ParamDefsC::getParams().getFloat( ParamDefsC::SEARCH_D ),
			ParamDefsC::getParams().getFloat( ParamDefsC::TOLERANCE ),
			ParamDefsC::getParams().getFloat( ParamDefsC::VOLUME_RATIO ),
			ParamDefsC::getParams().getFloat( ParamDefsC::GAMMA )
		);
				
		// set search direction
		assert( (pInnerMesh != NULL)  || (pOuterMesh != NULL) );
		int searchDirection = 0;		
		if ( pInnerMesh && pOuterMesh )
			searchDirection = 0;
		else if ( pInnerMesh )
			searchDirection = 1;
		else
			searchDirection = -1;

		coutVerbose("OK.\n");

		// do the funky stuff
		coutNormal("Minimizing...");

		const DSMC::ResultsC & rResults = dsm.minimize(
			&dsmParams,
			ParamDefsC::getParams().getInt( ParamDefsC::MAX_ROUNDS ),
			searchDirection);

		coutNormal("OK.\n");
		
/***************************************************************************
 * 
 * Save results
 * 
 ***************************************************************************/

		coutNormal("Saving results to \'");
		coutNormal(
			ParamDefsC::getParams().getString( ParamDefsC::OUTPUT_MESH ));
		coutNormal("\'...");
		CallbackHandlerC::save_mesh(
			rResults,
			ParamDefsC::getParams().getString( ParamDefsC::OUTPUT_MESH ));
		coutNormal("OK.\n");

		//save the final result also in intermediate files
		if ( ! myCallback.savedThisRound() )
			myCallback.saveIntermediate( rResults );

/***************************************************************************
 * 
 * Free allocated resources
 * 
 ***************************************************************************/

		if ( pOuterMesh ) delete pOuterMesh;
		if ( pInnerMesh ) delete pInnerMesh;
		if ( pImage ) delete pImage;

		coutVerbose("Bye!\n");

		return 0;

/***************************************************************************
 * 
 * Happily done!
 * 
 ***************************************************************************/
		

	} catch ( ExceptionC & e ) {
		cerr << e.what() << endl;	
	}

	} catch ( std::exception & e ) {
		cerr << e.what() << endl;	
	}

	} catch (...) {
		cerr << "Unknown exception." << endl;	
	}

	return -1;
	
} // main()


