/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "AnalyzeImageLoaderC.h"
#include "RAWImageLoaderC.h"
#include "ByteSwapper.hpp"
#include "ImageC.h"
#include "dbh.h"
#include <iolib/LoaderExceptionC.h>
#include <blitz/tinyvec.h>
#include <fstream>
#include <string>

using namespace imagelib;
using namespace std;
using namespace blitz;
using namespace iolib;

AnalyzeImageLoaderC & AnalyzeImageLoaderC::getInstance()
{
	static AnalyzeImageLoaderC * pInstance = NULL;
	if( ! pInstance )
		pInstance = new AnalyzeImageLoaderC;
	return *pInstance;
}

AnalyzeImageLoaderC::~AnalyzeImageLoaderC() throw()
{
}

const std::string & AnalyzeImageLoaderC::getDescription() const throw()
{
	static string desc("Analyze (tm) 7.5");
	return desc;
}

bool AnalyzeImageLoaderC::accepts( const std::string & rFileName ) const throw()
{
	return accepts(
		rFileName,
		ImageI::ParametersC(
			TinyVector<int,3>(0,0,0), // dimensions and datatype are ignore
			ImageI::FLOAT,            // by the loader.
			false )); // do not swap bytes by default.
}

ImageI::ParametersC AnalyzeImageLoaderC::loadHeader(std::istream & rStream)
{
	// read header
	dsr analHeader;
	rStream.read( reinterpret_cast<char *>(&analHeader), sizeof(dsr) );
	
	bool swab = false;
	// check header validity
	if (
		analHeader.dime.datatype != 2 &&	// uint8
		analHeader.dime.datatype != 4 &&	// int16
		analHeader.dime.datatype != 8 &&	// int32
		analHeader.dime.datatype != 16 &&	// float32
		analHeader.dime.datatype != 64 )	// float64
	{
		// try to swap bytes if failed
		swab = true;
        ByteSwapper::swapByteOrder(&analHeader.dime.datatype);
        ByteSwapper::swapByteOrder(&analHeader.dime.dim[1]);
        ByteSwapper::swapByteOrder(&analHeader.dime.dim[2]);
        ByteSwapper::swapByteOrder(&analHeader.dime.dim[3]);

		// check again
		if (
			analHeader.dime.datatype != 2 &&	// uint8
			analHeader.dime.datatype != 4 &&	// int16
			analHeader.dime.datatype != 8 &&	// int32
			analHeader.dime.datatype != 16 &&	// float32
			analHeader.dime.datatype != 64 )	// float64
		{
			// failed again!
			throw(LoaderExceptionC("Invalid \'.hdr\' file"));
		}
	}
	
	///@todo Allow vox_offset
	if ( analHeader.dime.vox_offset != 0 )
		throw ( iolib::LoaderExceptionC("Unsupported Analyze file.") );
	
	ImageI::DataTypeE dataType;
//	unsigned int precision;
	switch ( analHeader.dime.datatype )
	{
		case 2 :
			dataType = ImageI::UINT;
//			precision = 1;
			break;
		case 4 :
			dataType = ImageI::INT;
//			precision = 2;
			break;
		case 8 :
			dataType = ImageI::INT;
//			precision = 4;
			break;
		case 16 :
			dataType = ImageI::FLOAT;
//			precision = 4;
			break;
		case 64 :
			dataType = ImageI::FLOAT;
//			precision = 8;
			break;
		default:
			assert(false);
	}
	
	return ImageI::ParametersC(
		TinyVector<int,3>(
			analHeader.dime.dim[1],
			analHeader.dime.dim[2],
			analHeader.dime.dim[3]),
		dataType,
		swab );
}

bool AnalyzeImageLoaderC::accepts(
	const std::string & rFileName,
	const ImageI::ParametersC & rParams ) const throw()
{
	// check that header file is readable
	ifstream hdrFile( (rFileName + ".hdr").c_str(), ios::binary );
	if ( ! hdrFile.good() )
		return false;

	// check that image file is readable
	ifstream imgFile( (rFileName + ".img").c_str(), ios::binary );
	if ( ! imgFile.good() )
		return false;
	imgFile.close();
	
	loadHeader( hdrFile ); // this will throw on failure
	hdrFile.close();
	
	return true;	
}

ImageI * AnalyzeImageLoaderC::load( const std::string & rFileName, const ImageI::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	// check that header file is readable
	ifstream hdrFile( (rFileName + ".hdr").c_str(), ios::binary );
	if ( ! hdrFile.good() )
		return false;
		
	ImageI::ParametersC paramsFromHeader( loadHeader( hdrFile ) );
	hdrFile.close();
	
	ImageI::ParametersC finalParams(
		paramsFromHeader.getDimensions(),
		paramsFromHeader.getDataType(),
		paramsFromHeader.swapBytes() != rParams.swapBytes() );
		
	return RAWImageLoaderC::getInstance().load(
		(rFileName + ".img").c_str(),
		finalParams );	
}
	
ImageI * AnalyzeImageLoaderC::load( const std::string & rFileName ) const
	throw( iolib::ExceptionC )
{
	return load(
		rFileName,
		ImageI::ParametersC(
			TinyVector<int,3>(0,0,0), // dimensions and datatype are ignored
			ImageI::FLOAT,            // by the loader.
			false )); // do not swap bytes by default.
}
	
void AnalyzeImageLoaderC::save( const ImageI * pProduct, const std::string & rFileName, const ImageI::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	cerr << "AnalyzeImageLoaderC::save() - Not Implemented!" << endl;
}
	
void AnalyzeImageLoaderC::save( const ImageI * pProduct, const std::string & rFileName ) const
	throw( iolib::ExceptionC )
{
	//throw iolib::ExceptionC("Could not save raw voxel data; parameters required.");
	cerr << "AnalyzeImageLoaderC::save() - Not Implemented!" << endl;
}
	
AnalyzeImageLoaderC::AnalyzeImageLoaderC() throw()
{
}
