/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IMAGELIB_IMAGEI_H
#define IMAGELIB_IMAGEI_H

//#include "typedefs.h"
#include <blitz/array.h>
#include <blitz/tinyvec.h>
#include <iostream>

namespace imagelib
{
	/**A base class for voxel images.
	 * Basically just a wrapper for a Blitz++ Array.
	 */
	class ImageI
	{
	public:

		/**Voxel data types.
		 * For file loading.
		 */
		enum DataTypeE
		{
			FLOAT = 1,
			INT,
			UINT
		};			
		
		/**Parameters needed in image file loading.
		 */
		class ParametersC
		{
		public:
		
			/**A constructor.
			 * @param rDimensions Image dimensions (x,y,z)
			 * @param dataType Data type.
			 * @param swap @a true if byte order should be swapped
			 */
			ParametersC(
				const blitz::TinyVector<int,3> & rDimensions,
				DataTypeE dataType,
				bool swap ) throw()
			: m_dimensions(rDimensions),
			  m_dataType(dataType),
			  m_swapBytes(swap)
			{
			};
	
			/**A non-virtual destructor.
			 */
			~ParametersC() throw() {};
	
			/**Return dimensions parameter.
			 */
			const blitz::TinyVector<int,3> & getDimensions() const throw()
			{
				return m_dimensions;
			};
	
			/**Return data type.
			 */
			DataTypeE getDataType() const throw() {
				return m_dataType;
			};
			
			/**Return byteswap flag.
			 */
			bool swapBytes() const throw() {
				return m_swapBytes;
			};
			
		private:
		
			/**Dimensions.
			 */
			blitz::TinyVector<int,3> m_dimensions;
			
			/**Data type.
			 */
			DataTypeE m_dataType;
			
			/**SwaB flag.
			 */
			bool m_swapBytes;
		}; //class ParametersC
		
		/**A virtual destructor.
		 */
		virtual ~ImageI() throw() {};

		/**Return image dimensions.
		 */
		virtual const blitz::TinyVector<int,3> getDimensions()
			const throw () = 0;
		
		/**Normalize image data.
		 */
		virtual void normalize() throw () = 0;

		/**Return image data as a Blitz++ Array.
		 */
		virtual blitz::Array<float,3> * getImageData() throw() = 0;

		/**Return image data as a Blitz++ Array.
		 */
		virtual const blitz::Array<float,3> * getImageData()
			const throw () = 0;
			
	}; // class ImageI
}; // namespace imagelib

#endif //IMAGELIB_IMAGEI_H
