/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "RAWImageLoaderC.h"
#include "ByteSwapper.hpp"
#include "ImageC.h"
#include <iolib/LoaderExceptionC.h>
//#include <blitz/tinyvec.h>
#include <fstream>
#include <string>

using namespace imagelib;
using namespace std;
using namespace blitz;

// currently supported data types:
// float 32,64
// (u)int 8,16,32
//

RAWImageLoaderC & RAWImageLoaderC::getInstance()
{
	static RAWImageLoaderC * pInstance = NULL;
	if( ! pInstance )
		pInstance = new RAWImageLoaderC;
	return *pInstance;
}

RAWImageLoaderC::~RAWImageLoaderC() throw()
{
}

const std::string & RAWImageLoaderC::getDescription() const throw()
{
	static string desc("Raw voxel data");
	return desc;
}

bool RAWImageLoaderC::accepts( const std::string & rFileName ) const throw()
{
	return false;
}

bool RAWImageLoaderC::accepts( const std::string & rFileName, const ImageI::ParametersC & rParams ) const throw()
{
	TinyVector<int,3> dim( rParams.getDimensions() ); //get desired image dimensions

	unsigned int uiSize( dim(0) * dim(1) * dim(2) ); // calculate image size in voxels

	ifstream rawFile( rFileName.c_str() ); // get file size
	rawFile.seekg( 0,ios::end );
	unsigned int uiFileSize( rawFile.tellg() );
	rawFile.close();

	if( (uiFileSize % uiSize ) != 0 ) // check if file size matches
		return false;
		
	unsigned int uiPrecision( uiFileSize / uiSize ); // calculate data precision

	switch( rParams.getDataType() ) // check if precision is supported
	{
		case ImageI::FLOAT:
			return uiPrecision == 4 || uiPrecision == 8 || uiPrecision == 10;
		case ImageI::INT:
		case ImageI::UINT:
			return uiPrecision == 8 || uiPrecision == 16 || uiPrecision == 32;
		default :
			return false;
	}
}

template<typename T, int N>
void RAWImageLoaderC::read_data( float *pData, istream & rStream, unsigned int uiCount, bool bSwap )
{
	T dataItem = 0;
	char * pDataItem = (char *)(&dataItem);
	while( uiCount-- )
	{
		rStream.read(pDataItem,N);
		if( bSwap )
			ByteSwapper::swapByteOrder(&dataItem);
		(*pData++) = (float)dataItem;
	}	
}

ImageI * RAWImageLoaderC::load( const std::string & rFileName, const ImageI::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	//TODO: catch bad_alloc
	TinyVector<int,3> dim( rParams.getDimensions() );
	unsigned int uiSize( dim(0) * dim(1) * dim(2) );
	float * pData = new float[uiSize]; // will be deallocated by ~Array<>
	
	ifstream rawFile( rFileName.c_str(), ios::in | ios::binary );
	rawFile.seekg( 0, ios::end );
	unsigned int uiFileSize( rawFile.tellg() );
	rawFile.seekg( 0, ios::beg );

	if( (uiFileSize % uiSize ) != 0 ) // check if file size matches
		throw iolib::LoaderExceptionC("Could not load raw voxel data: file size mismatch");
		
	unsigned int uiPrecision( uiFileSize / uiSize ); // calculate data precision
	
	switch( rParams.getDataType() )
	{
/*		float f32Dummy;
		double f64Dummy;
		unsigned char ui8Dummy;
		unsigned short ui16Dummy;
		unsigned int ui32Dummy;
		signed char i8Dummy;
		signed short i16Dummy;
		signed int i32Dummy;*/
		
		case ImageI::FLOAT :
		{
			switch( uiPrecision )
			{
				case 4 :				
					read_data<float,4>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 8 :
					read_data<double,8>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 10 :
					read_data<long double,10>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				default :
					throw iolib::LoaderExceptionC(
						"Could not load raw voxel data: unsupported floating point precision");
			}
			break;
		}

		case ImageI::INT :
		{
			switch( uiPrecision )
			{
				case 1 :
					read_data<signed char,1>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 2 :
					read_data<signed short,2>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 4 :
					read_data<signed long,4>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				default :
					throw iolib::LoaderExceptionC(
						"Could not load raw voxel data: unsupported uint precision");
			}
			break;
		}

		case ImageI::UINT :
		{
			switch( uiPrecision )
			{
				case 1 :
					read_data<unsigned char,1>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 2 :
					read_data<unsigned short,2>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				case 4 :
					read_data<unsigned long,4>( pData, rawFile, uiSize, rParams.swapBytes() );
					break;
				default :
					throw iolib::LoaderExceptionC(
						"Could not load raw voxel data: unsupported integer precision");
			}
			break;
		}

		default:
		{
			throw iolib::LoaderExceptionC(
				"Could not load raw voxel data: unrecognized data type");
			break;
		}
	}
		
//	Array<float,3> * pArray = 
//		new Array<float,3>( pData, dim, deleteDataWhenDone );
		// will be deallocated by ~ImageC

	Array<float,3> * pArray = new Array<float,3>(dim);

	for(int z = 0; z < dim(2); ++z)
	for(int y = 0; y < dim(1); ++y)
	for(int x = 0; x < dim(0); ++x)
		(*pArray)(x,y,z) = pData[z*dim(0)*dim(1) + y*dim(0) + x];
		
	delete pData;
		
	ImageI * pImage = new ImageC(pArray);
	return pImage;
}
	
ImageI * RAWImageLoaderC::load( const std::string & rFileName ) const
	throw( iolib::ExceptionC )
{
	throw iolib::ExceptionC("Could not load raw voxel data; parameters required.");
}
	
void RAWImageLoaderC::save( const ImageI * pProduct, const std::string & rFileName, const ImageI::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	cerr << "RAWImageLoaderC::save() - Not Implemented!" << endl;
}
	
void RAWImageLoaderC::save( const ImageI * pProduct, const std::string & rFileName ) const
	throw( iolib::ExceptionC )
{
	//throw iolib::ExceptionC("Could not save raw voxel data; parameters required.");
	cerr << "RAWImageLoaderC::save() - Not Implemented!" << endl;
}
	
RAWImageLoaderC::RAWImageLoaderC() throw()
{
}
