/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IOLIB_FACTORYT_H
#define IOLIB_FACTORYT_H

#include "LoaderT.h"
#include <string>
#include <list>

namespace iolib
{
	
	/** A template for a factory.
	 * An instance of this template is a factory that is capable of creating
	 * objects of type ProductType. Instances of this template are
	 * singletons.
	 * @param ProductType Type of the objects to be created, e.g. MeshI
	 * @see LoaderT
	 * @see MeshFactoryC
	 * @see ImageFactoryC
	 */
	template<class ProductType>
	class FactoryT
	{
	protected:

		/** A protected constructor.
		 * getInstance() should be used to create factory objects.
		 */
		FactoryT() throw();
		
		/** List of registered loaders.
		 */
		std::list< const LoaderT< ProductType > * > m_loaders;
				
	public:
	
		/** A virtual destructor for safety reasons
		 */
		virtual ~FactoryT() throw();
	
		/** Get factory instance.
		 * @return A reference to the singleton instance of the factory
		 */
		static FactoryT<ProductType> & getInstance() throw( ExceptionC );
				
		/** Register a new file loader.
		 * This method registers a new file loader to the factory.
		 * @param rLoader A reference to a file loader object.
		 */
		void addLoader( const LoaderT<ProductType> & rLoader )
			throw ( ExceptionC );
		
		/** Creates a product from a file.
		 * This method creates a product from a file if an appropriate file
		 * loader is registered to the factory. If file loading fails, NULL
		 * pointer is returned.
		 * @param rFileName The name of the file to be loaded
		 * @return A pointer to the newly created object. It is in the
		 * caller's responsibility to deallocate the object.
		 */
		ProductType * create( const std::string & rFileName )
			throw ( ExceptionC );
		
		/** Creates a product from a file with additional parameters.
		 * This method creates a product from a file if an appropriate file
		 * loader is registered to the factory. If file loading fails, NULL
		 * pointer is returned.
		 * @param rFileName The name of the file to be loaded
		 * @param rParams Additional parameters neede in file loading
		 * @return A pointer to the newly created object. It is in the
		 * caller's responsibility to deallocate the object.
		 */
		ProductType * create(
			const std::string & rFileName,
			const typename ProductType::ParametersC & rParams )
		throw( ExceptionC );
		
	};// template class FactoryT
	
	template<class ProductType>
	FactoryT<ProductType>::FactoryT() throw()
	{
	};

	template<class ProductType>
	FactoryT<ProductType>::~FactoryT() throw()
	{
	};
	
	template<class ProductType>
	FactoryT<ProductType> & FactoryT<ProductType>::getInstance()
		throw ( ExceptionC )
	{
		static FactoryT<ProductType> * pInstance = NULL;
		
		try
		{
			if( pInstance == NULL )
				pInstance = new FactoryT<ProductType>();
			return *pInstance;
		}
		catch ( std::bad_alloc & )
		{
			throw ExceptionC( "Factory instance could not be allocated." );
		}
	};
	
	template<class ProductType>
	void FactoryT<ProductType>::addLoader(
		const LoaderT<ProductType> & rLoader ) throw ( ExceptionC )
	{
		m_loaders.push_front( &rLoader );
	};
	
	/** This method tries to find a loader that accepts the given file
	 * and returns NULL if none is found.
	 */
	template<class ProductType>
	ProductType * FactoryT<ProductType>::create(
		const std::string & rFileName ) throw ( ExceptionC )
	{
		ProductType * pProduct = NULL;
		typename std::list< const LoaderT< ProductType > * > ::const_iterator
			itLoader = m_loaders.begin();
		while ( itLoader != m_loaders.end() )
		{
			if ( (*itLoader)->accepts( rFileName ) )
			{
				try
				{
					pProduct = ( *itLoader )->load( rFileName );
					if( pProduct != NULL)
						return pProduct;
				}
				catch ( ExceptionC & )
				{
					;	// It is perfectly acceptable for a loader to fail
						// with an exception.
				}
			}
			++itLoader;
		}
		return NULL;
	};
	
	
	/** This method tries to find a loader that accepts the given file
	 * and loader parameters, and returns NULL if none is found.
	 */
	template<class ProductType>
	ProductType * FactoryT<ProductType>::create(
		const std::string & rFileName,
		const typename ProductType::ParametersC & rParams )
	throw ( ExceptionC )
	{
		ProductType * pProduct = NULL;
		typename std::list< const LoaderT< ProductType > * > ::const_iterator
			itLoader = m_loaders.begin();
		while ( itLoader != m_loaders.end() )
		{
			if ( (*itLoader)->accepts( rFileName, rParams ) )
			{
				try
				{
					pProduct = ( *itLoader )->load( rFileName, rParams );
					if( pProduct != NULL)
						return pProduct;
				}
				catch ( ExceptionC & )
				{
					;	// It is perfectly acceptable for a loader to fail
						// with an exception.
				}
			}
			++itLoader;
		}
		return NULL;
	};
	
};	// namespace iolib

#endif //IOLIB_FACTORYT_H
