/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IOLIB_LOADERT_H
#define IOLIB_LOADERT_H

#include <string>
#include <vector>
#include "ExceptionC.h"

namespace iolib
{		
	/** A template for file loaders.
	 *  Instances of this template are abstract base classes for file loaders.
	 *  @param ProductType the class of the product to be loaded (e.g. MeshI)
	 *  @see FactoryT
	 *  @see MeshLoaderI
	 *  @see ImageLoaderI
	 */
	template<class ProductType>
	class LoaderT
	{
	public:

		/** A virtual destructor for safety reasons.
		 */
		virtual ~LoaderT() throw () { };
		
		/** File format description.
		 * An implementation of this method should return a short description
		 * of the file format(s) that the loader is able to handle.
		 * For example, a VRML mesh loader could return "VRML 1.0". It should
		 * be noted that the value returned by this method should be used for
		 * UI purposes only, and should not be used to recognize the type of
		 * the loader.
		 * @return File format description string
		 */
		virtual const std::string & getDescription() const throw () = 0;
		
		/** File format test.
		 * An implementation of this method should return true if the loader
		 * thinks it is capable of loading the specified file.
		 * It is however possible for the load() method to fail even if this
		 * method returned true.
		 * @param rFileName The name of the file to be checked
		 * @return true if loading is possible, false otherwise
		 */
		virtual bool accepts( const std::string & rFileName ) const
			throw () = 0;

		/** File format test.
		 * An implementation of this method should return true if the loader
		 * thinks it is capable of loading the specified file using the
		 * specified loader parameters.
		 * It is however possible for the load() method to fail even if this
		 * method returned true.
		 * @param rFileName The name of the file to be checked
		 * @param rParams Additional parameters needed in file loading
		 * @return true if loading is possible, false otherwise
		 */
		virtual bool accepts(
			const std::string & rFileName,
			const typename ProductType::ParametersC & rParams
		) const throw () = 0;
		
		/** Creates a product from a file.
		 * @param rFileName The name of the input file, e.g. 'inner.wrl'
		 * @param rParams Additional parameters needed in file loading
		 * @return A pointer to the newly created object. It is in the
		 * caller's responsibility to deallocate the object.
		 */
		virtual ProductType * load (
			const std::string & rFileName,
			const typename ProductType::ParametersC & rParams
		) const throw ( ExceptionC ) = 0;

		/** Creates a product from a file.
		 * @param rFileName The name of the input file, e.g. 'inner.wrl'
		 * @return A pointer to the newly created object. It is in the
		 * caller's responsibility to deallocate the object.
		 */
		virtual ProductType * load ( const std::string & rFileName ) const
			throw ( ExceptionC ) = 0;
		
		/** Saves a product to a file.
		 * @param pProduct A pointer to the object to be saved
		 * @param rFileName The name of the output file, e.g. 'result.wrl'
		 * @param rParams Additional parameters needed in file saving
		 */
		virtual void save(
			const ProductType * pProduct,
			const std::string & rFileName,
			const typename ProductType::ParametersC & rParams
		) const throw ( ExceptionC ) = 0;

		/** Saves a product to a file.
		 * @param pProduct A pointer to the object to be saved
		 * @param rFileName The name of the output file, e.g. 'result.wrl'
		 */
		virtual void save(
			const ProductType * pProduct,
			const std::string & rFileName ) const throw( ExceptionC ) = 0;

	}; // template class LoaderT
	
};	// namespace iolib

#endif //IOLIB_LOADERT_H
