/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MESHLIB_FACEC_H
#define MESHLIB_FACEC_H

#include "typedefs.h"
#include <vector>

namespace meshlib
{
	// forward declarations
	class MexelC;
	
	/**
	 * Represents a face (of a mesh).
	 * This class encapsulates the properties of a single face of a mesh.
	 * A face is basically an ordered list of mexels (vertices).
	 * @see MeshC
	 * @see MexelC
	 * @todo It would be better if MeshC would act as a factory for FaceC
	 * objects.
	 */
	class FaceC
	{
	private:
	
		/** A vector of MexelC pointers.
		 */
		std::vector<MexelC*> m_mexelPointers;
		
	public:

		/**The default constructor.
		 * Mexel list is initialized to empty.
		 */
		FaceC() throw ()
		{
		};
		
		/**A copy constructor.
		 */
		FaceC( const FaceC & rFace ) throw ()
			: m_mexelPointers( rFace.m_mexelPointers )
		{
		};
		
		/** The destructor.
		 * As the destructor is not virtual, this class is not designed
		 * for subclassing.
		 */
		~FaceC() throw ()
		{
		};
				
		/**
		 * Returns the number of mexels in the face.
		 * @return The number of mexels.
		 */
		int getMexelCount() const throw ()
		{
			return m_mexelPointers.size();
		};
		
		/**Returns a const iterator to the first mexel.
		 * @return A const iterator
		 */
		MexelConstIterator mexelBegin() const throw ()
		{
			return m_mexelPointers.begin();
		};

		/**Returns a const iterator to the end of mexel list.
		 * Note that 'end' means 'one beyond last'.
		 * @return A const iterator
		 */
		MexelConstIterator mexelEnd() const throw ()
		{
			return m_mexelPointers.end();
		};

		/**Returns an iterator to the first mexel.
		 * @return An iterator
		 */
		MexelIterator mexelBegin() throw ()
		{
			return m_mexelPointers.begin();
		};

		/**Returns an iterator to the end of mexel list.
		 * Note that 'end' means 'one beyond last'.
		 * @return An iterator
		 */
		MexelIterator mexelEnd() throw ()
		{
			return m_mexelPointers.end();
		};
		
		/**Adds a mexel pointer to the end of mexel list.
		 * Null-pointer check is done only in debug mode.
		 * @param pMexel A pointer to a mexel
		 */
		void addMexel( MexelC * pMexel ) throw ()
		{
			assert( pMexel != NULL );
			m_mexelPointers.push_back( pMexel );
		};
	};	// class FaceC
	
}; // namespace meshlib

#endif //MESHLIB_FACEC_H
