/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "MeshC.h"
#include "MexelC.h"
#include "FaceC.h"

#include <blitz/tinyvec-et.h>

using namespace meshlib;
using namespace std;

MeshC::MeshC() throw()
{
}

bool MeshC::isSimplex() const throw ()
{
	MexelConstIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		if ( (*itMex)->getNeighborCount() != 3 )
			return false;
		++itMex;
	}
	return true;
}

MeshC::MeshC( const MeshC & rMesh ) throw ()
 : m_refPoint(rMesh.getReferencePoint()),
 m_averageArea(rMesh.getAverageArea()),
 m_volume(rMesh.getVolume()),
 m_volumeMetric(rMesh.getVolumeMetric()),
 m_moved(rMesh.hasMoved())
{
	//copy mexels
	MexelConstIterator itMex = rMesh.mexelBegin();
	while ( itMex != rMesh.mexelEnd() )
	{
		addMexel( MexelC( (*itMex)->getLocation() ) );
		++itMex;
	}
	
	//form neighbourhood relation
	itMex = rMesh.mexelBegin();
	MexelIterator itMex2 = mexelBegin();
	while ( itMex != rMesh.mexelEnd() )
	{
		(*itMex2)->addNeighbor(
			getMexel( (*itMex)->getNeighbor(0)->getIndex() ));
		(*itMex2)->addNeighbor(
			getMexel( (*itMex)->getNeighbor(1)->getIndex() ));
		(*itMex2)->addNeighbor(
			getMexel( (*itMex)->getNeighbor(2)->getIndex() ));
		++itMex;
		++itMex2;
	}

	//copy faces
	FaceConstIterator itFace = rMesh.faceBegin();
	while ( itFace != rMesh.faceEnd() )
	{
		FaceC face;
		MexelConstIterator itPoint = (*itFace)->mexelBegin();
		while ( itPoint != (*itFace)->mexelEnd() )
		{
			face.addMexel( getMexel( (*itPoint)->getIndex() ) );
			++itPoint;
		}
		addFace(face);
		++itFace;
	}	
}

void MeshC::storeMexelLocations(
	std::vector<MLVector> & rLocs ) const throw ()
{
	rLocs.clear();
	MexelConstIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		rLocs.push_back((*itMex)->getLocation());
		++itMex;
	}	
}

void MeshC::restoreMexelLocations(
	const std::vector<MLVector> & rLocs ) throw (iolib::ExceptionC)
{
	assert( rLocs.size() == getMexelCount() );

	MexelConstIterator itMex = mexelBegin();
	std::vector<MLVector>::const_iterator itLoc = rLocs.begin();
	while ( itMex != mexelEnd() )
	{
		(*itMex)->setLocation(*itLoc);
		++itMex;
		++itLoc;
	}	
}


MeshC::~MeshC() throw()
{
	FaceIterator itFace = faceBegin();
	while ( itFace != faceEnd() )
	{
		delete *itFace;
		++itFace;
	}
	
	MexelIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		delete *itMex;
		++itMex;
	}
}

unsigned int MeshC::getMexelCount() const throw()
{
	return m_mexels.size();
}

void MeshC::triangulate()
{
//	cout << "*** MeshC::triangulate() - Not implemented! ***" << endl;
//	assert(false);

	vector<FaceC*> oldFaces(m_faces);
	
	m_faces.clear();
	FaceConstIterator itFace( oldFaces.begin() );
	while ( itFace != oldFaces.end() )
	{
		MLVector faceCenter(0.0, 0.0, 0.0);
		unsigned int pointCount = 0;
		MexelIterator itPoint( (*itFace)->mexelBegin() );
		while ( itPoint != (*itFace)->mexelEnd() )
		{
			faceCenter += (*itPoint)->getLocation();
			++itPoint;
			++pointCount;
		}
		faceCenter /= ((MLFloat)pointCount);
		
		MexelC * pCenterMex = addMexel( MexelC(faceCenter) );
		
		itPoint = (*itFace)->mexelBegin();
		MexelC * pMex = (*itPoint);
		MexelC * pPrevMex = NULL;
		MexelC * pFirstMex = pMex;
		++itPoint;
		while ( itPoint != (*itFace)->mexelEnd() )
		{
			pPrevMex = pMex;
			pMex = *itPoint;
	
			FaceC newFace;
			newFace.addMexel(pCenterMex);
			newFace.addMexel(pPrevMex);
			newFace.addMexel(pMex);
			addFace(newFace);
	
			++itPoint;
		}
		FaceC lastNewFace;
		lastNewFace.addMexel(pCenterMex);
		lastNewFace.addMexel(pMex);
		lastNewFace.addMexel(pFirstMex);
		addFace(lastNewFace);
		
		delete *itFace; // free old faces
		++itFace;
	}
}

void MeshC::scale( const MLVector & rScale )
{
	MexelIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		(*itMex)->setLocation( rScale * ((*itMex)->getLocation()) );
		++itMex;
	}
}

void MeshC::translate( const MLVector & rOffset )
{
	MexelIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		(*itMex)->setLocation( rOffset + ((*itMex)->getLocation()) );
		++itMex;
	}
}

MLFloat MeshC::getAverageArea() const throw ()
{
	return m_averageArea;
}

MLFloat MeshC::getVolume() const throw ()
{
	///@todo
//	assert(false);
	return 1.0;
}

MLFloat MeshC::getVolumeMetric() const throw ()
{
	///@todo
	MLFloat vm = 0.0;
	MexelConstIterator itMex = mexelBegin();
	while ( itMex != mexelEnd() )
	{
		vm += sum( abs( (*itMex)->getLocation() ) );
		++itMex;
	}
	return vm;
}

void MeshC::calcAverageArea() throw ()
{
	m_averageArea = 0.0;
	FaceConstIterator itFace = faceBegin();
	while ( itFace != faceEnd() )
	{
		MexelConstIterator itMex = (*itFace)->mexelBegin();
		MLVector faceOrigin( (*itMex)->getLocation() );
		++itMex;
		MLVector sideA(0.0,0.0,0.0);
		MLVector sideB( (*itMex)->getLocation() - faceOrigin );
		++itMex;
		while ( itMex != (*itFace)->mexelEnd() )
		{
			sideA = sideB;
			sideB = (*itMex)->getLocation() - faceOrigin;
			MLVector aCrossB = cross(sideA,sideB);
			// 20.10.04
			// m_averageArea += sqrt(sum(aCrossB*aCrossB));
			m_averageArea += (0.5*(sqrt(sum(aCrossB*aCrossB))));


			++itMex;
		}

		++itFace;
	}

	m_averageArea /= getFaceCount();
}

void MeshC::calcVolume() throw ()
{
	///@todo
}

void MeshC::calcVolumeMetric() throw ()
{
	///@todo
}

unsigned int MeshC::getFaceCount() const throw()
{
	return m_faces.size();
}

void MeshC::setReferencePoint( const MLVector & rPoint ) throw()
{
	m_refPoint = rPoint;
}

const MLVector & MeshC::getReferencePoint() const throw()
{
	return m_refPoint;
}

MexelIterator MeshC::mexelBegin() throw()
{
	return m_mexels.begin();
}

MexelConstIterator MeshC::mexelBegin() const throw()
{
	return m_mexels.begin();
}

MexelIterator MeshC::mexelEnd() throw()
{
	return m_mexels.end();
}

MexelConstIterator MeshC::mexelEnd() const throw()
{
	return m_mexels.end();
}

FaceIterator MeshC::faceBegin() throw()
{
	return m_faces.begin();
}

FaceConstIterator MeshC::faceBegin() const throw()
{
	return m_faces.begin();
}

FaceIterator MeshC::faceEnd() throw()
{
	return m_faces.end();
}

FaceConstIterator MeshC::faceEnd() const throw()
{
	return m_faces.end();
}

MexelC * MeshC::addMexel( const MexelC& rMexel ) throw( iolib::MemoryAllocationExceptionC )
{
	MexelC * pMex = new MexelC(rMexel, m_mexels.size());
//	pMex->setIndex(m_mexels.size());
	m_mexels.push_back(pMex);
	return pMex;
}

FaceC * MeshC::addFace( const FaceC& rFace ) throw( iolib::MemoryAllocationExceptionC )
{
	FaceC * pFace = new FaceC(rFace);
	m_faces.push_back(pFace);
	return pFace;
}

