/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MESHLIB_MESHC_H
#define MESHLIB_MESHC_H

#include "typedefs.h"

#include <iolib/MemoryAllocationExceptionC.h>

#include <vector>
#include <string>

namespace meshlib
{
	/**Represent a (simplex) 3D mesh.
	 */
	class MeshC
	{
	public:	

		/**Parameters needed for mesh loading/saving.
		 */
		class ParametersC
		{
		public:

			/**Constructor for mesh saving parameters.
			 * @param rComment Mesh file comment.
			 * This string will be included as a comment
			 * in the saved mesh file (if the file format allows)
			 */
			ParametersC( const std::string & rComment ) throw ()
				: m_comment( rComment ), m_refPoint(0,0,0) {};
				
			/**Constructor for mesh loading parameters.
			 * @param rRefPoint Reference point to be used.
			 * This information will be used by the mesh file loader to
			 * set the reference.
			 */
			ParametersC( const MLVector & rRefPoint ) throw ()
				: m_comment(""), m_refPoint(rRefPoint) {};

			/**A non-virtual destructor.
			 * This class is not designed for subclassing.
			 */
			~ParametersC() throw () {};
			
			/**Return the mesh file comment.
			 *@return Mesh file comment
			 */
			const std::string & getComment() const throw ()
			{
				return m_comment;
			};
			
			/**Return the reference point.
			 * @return Reference point
			 */
			const MLVector & getReferencePoint() const throw ()
			{
				return m_refPoint;
			};
			
		private:
		
			/**Mesh file comment.
			 */
			std::string m_comment;
			
			/**Reference point.
			 */
			MLVector m_refPoint;
			
		}; // class ParametersC
	
		/**The default constructor.
		 * Constructs an empty mesh.
		 */
		MeshC() throw();
		
		/**Copy constructor.
		 * All mexels and faces are copied.
		 */
		MeshC( const MeshC & rMesh ) throw();
		
		/**A non-virtual destructor.
		 * MeshC is not designed for subclassing.
		 */
		~MeshC() throw();
		
		/**Set reference point.
		 * @param rPoint The reference point
		 */
		void setReferencePoint( const MLVector & rPoint ) throw();
		
		/**Return reference point.
		 * @return The reference point
		 */
		const MLVector & getReferencePoint() const throw();		

		/**Return mexel count.
		 * @return The number of mexels (vertices)
		 */
		unsigned int getMexelCount() const throw();
		
		/**Return face count.
		 * @return The number of faces (polygons)
		 */
		unsigned int getFaceCount() const throw();

		bool isSimplex() const throw ();			

		/**Triangulate the mesh.
		 * Note that after @a triangulate() the mesh will no longer be
		 * simplex, and the neighborhood relations are screwed.
		 */
		void triangulate();
		
		/**Scale the mesh.
		 * @param rScale Scale factors in (x,y,z) directions
		 */
		void scale( const MLVector & rScale );
		
		/**Translate the mesh.
		 * @param rOffset Offset in 3-space
		 */
		void translate( const MLVector & rOffset );
		
		/**Store mexel locations.
		 * Takes a snapshot of mexel locations. Any previous contents of
		 * @a rLocs will be cleared.
		 * @param rLocs A reference to a stl vector of locations in 3-space
		 * @see restoreMexelLocations
		 */
		void storeMexelLocations(
			std::vector<MLVector> & rLocs ) const throw ();

		/**Restore mexel locations.
		 * Restores mexel locations from a snapshot.
		 * If the number of locations in @a rLocs does not match the number
		 * of mexels in the mesh, an exception will be thrown.
		 * @param rLocs A reference to a stl vector of locations in 3-space
		 */
		void restoreMexelLocations( const std::vector<MLVector> & rLocs )
			throw (iolib::ExceptionC);

		/**Return the average surface area of the faces on the mesh.
		 * @return Average area
		 */
		MLFloat getAverageArea() const throw ();

		/**Return the volume of the mesh.
		 * @return Volume
		 */
		MLFloat getVolume() const throw ();

		/**Return a metric that approximates the volume of the mesh.
		 * The only thing that is guaranteed of the volume metric is
		 * that it is order-preserving:
		 * @a volume1 < @a volume2 implies @a metric1 < @a metric2 and
		 * @a volume1 > @a volume2 implies @a metric1 > @a metric2.
		 * @return Volume metric.
		 */
		MLFloat getVolumeMetric() const throw ();

		/**Return the state of the @a moved -flag.
		 * The moved-flag is used to track whether the mesh has changed.
		 * @see setMoved
		 */
		bool hasMoved() const throw () { return m_moved; };

		/**Set the state of the @a moved -flag.
		 * The moved-flag is used to track whether the mesh has changed.
		 * This is the only method call that affects the state of the flag.
		 * @param moved A new state for the flag.
		 * @see hasMoved
		 */
		void setMoved( bool moved = true ) throw () { m_moved = moved; };
		
		/**Update average area.
		 */
		void calcAverageArea() throw ();
		
		/**Update volume.
		 */
		void calcVolume() throw ();
		
		/**Update the volume metric.
		 */
		void calcVolumeMetric() throw ();

		/**Return an iterator to the first mexel.
		 */
		MexelIterator mexelBegin() throw();

		/**Return a const iterator to the first mexel.
		 */
		MexelConstIterator mexelBegin() const throw();

		/**Return an iterator to one beyond the last mexel.
		 */
		MexelIterator mexelEnd() throw();

		/**Return an iterator to one beyond the last mexel.
		 */
		MexelConstIterator mexelEnd() const throw();

		/**Return an iterator to the first face.
		 */
		FaceIterator faceBegin() throw();

		/**Return an iterator to the first face.
		 */
		FaceConstIterator faceBegin() const throw();

		/**Return an iterator to one beyond the last face.
		 */
		FaceIterator faceEnd() throw();

		/**Return an iterator to one beyond the last face.
		 */
		FaceConstIterator faceEnd() const throw();

		/**Return a mexel by index.
		 * @param index Mexel index (0...N-1)
		 */
		MexelC * getMexel( int index ) const throw()
		{
			return m_mexels.at(index);
		};
		
		/**Add a new mexel.
		 * A copy is made of @a rMexel.
		 * @param rMexel A mexel whose properties are to be copied.
		 * @return A pointer to the newly added mexel.
		 */
		MexelC * addMexel( const MexelC& rMexel )
			throw( iolib::MemoryAllocationExceptionC );
			
		/**Add a new face.
		 * A copy is made of @a rFace.
		 * @param rFace A face whose properties are copied.
		 * @return A pointer to the newly added face.
		 */
		FaceC * addFace( const FaceC& rFace )
			throw( iolib::MemoryAllocationExceptionC );
			
	private:

		/**Mexel list.
		 */
		std::vector<MexelC*> m_mexels;
		
		/**Face list.
		 */
		std::vector<FaceC*> m_faces;
		
		/**Reference point.
		 */
		MLVector m_refPoint;
		
		/**Cached average area.
		 */
		MLFloat m_averageArea;
		
		/**Cached volume.
		 */
		MLFloat m_volume;
		
		/**Cached volume metric.
		 */
		MLFloat m_volumeMetric;
		
		/**The @a moved -flag.
		 */
		bool m_moved;
		
	}; // class MeshC
}; // namespace meshlib

#endif //MESHLIB_MESHC_H
