/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MESHLIB_MEXELC_H
#define MESHLIB_MEXELC_H

#include "typedefs.h"
#include <vector>
#include <iostream>

using namespace std;

namespace meshlib
{	
	/**
	 * Represents a mexel (mesh element).
	 * This class encapsulates the properties of a single mexel (of a
	 * simplex mesh). This
	 * includes a location in 3-space, and an ordered list of neighboring
	 * mexels. In the case of a simplex mesh, the number of neighbors
	 * is always three.
	 */
	class MexelC
	{

	private:

		/**Location in 3-space.
		 */
		MLVector m_location;
		
		/**Neighborhood relation.
		 */
		MexelC * m_neighbors [ 3 ];
		
		/**Mexel index.
		 * @todo This is really only needed by MeshLoaderI::save().
		 * Try to make up an alternative strategy to find out mexel index.
		 */
		unsigned int m_index;
		
		//mutable bool m_moved;
		
	public:
		/**
		 * Default constructor.
		 * Location will be initialized to (0,0,0), and neighbor list will be empty.
		 */
		MexelC( unsigned int index = 0 ) throw ()
			:	m_location( 0.0, 0.0, 0.0 ),
				m_index( index )
		{
			m_neighbors[0] = NULL;
			m_neighbors[1] = NULL;
			m_neighbors[2] = NULL;
		};
		
		/**
		 * Constructor.
		 */
		MexelC( MLFloat x, MLFloat y, MLFloat z, unsigned int index = 0 )
			throw ()
			:	m_location( x, y, z ),
				m_index(index)
		{
			m_neighbors[0] = NULL;
			m_neighbors[1] = NULL;
			m_neighbors[2] = NULL;
		};
		
		/**
		 * Constructor.
		 * @param rLocation a location in 3-space
		 * @param index The mexel's index in the mesh
		 */
		MexelC( const MLVector & rLocation, unsigned int index = 0 )
			throw ()
			:	m_location( rLocation ),
				m_index( index )
		{
			m_neighbors[0] = NULL;
			m_neighbors[1] = NULL;
			m_neighbors[2] = NULL;
		};
		
		/**A copy constructor.
		 */
		MexelC( const MexelC & rMexel, unsigned int index = 0 ) throw ()
		:	m_location(rMexel.m_location),
			m_index(index)
		{
			m_neighbors[0] = rMexel.m_neighbors[0];
			m_neighbors[1] = rMexel.m_neighbors[1];
			m_neighbors[2] = rMexel.m_neighbors[2];
		};
		
		
		/**
		 * Destructor.
		 */
		~MexelC() throw ()
		{
		};

		
		/**Return the mexel's location in 3-space.
		 */
		const MLVector & getLocation() const throw ()
		{
			return m_location;
		};
		
		/**Set the mexel's location in 3-space.
		 */
		void setLocation( const MLVector & rLocation ) throw ()
		{
			m_location = rLocation;
		};
		
		/**Return the nth neighbor of the mexel.
		 * @param n Must be 0..2
		 */
		MexelC * getNeighbor( unsigned int n ) const throw ()
		{
			assert( n < 3 );
			return m_neighbors [ n ];
		};

		/**Adds a neighbor to the mexel's neighbor list.
		 */
		void addNeighbor( MexelC* pMexel ) throw ()
		{
			assert( pMexel != NULL );			
			int i = 0;
			while ( (m_neighbors [ i ] != NULL ) && i < 3 )
				++i;
			if ( i < 3 )
				m_neighbors [ i ] = pMexel;
/*			else
			{
				cout << "ERROR: input mesh is not simplex!" << endl;
				exit(-1);
			}*/
		};
		
		unsigned int getNeighborCount() const throw ()
		{
			unsigned int count = 0;
			for ( int i = 0; i < 3; ++i )
				if ( m_neighbors [ i ] != NULL )
					++count;
			return count;
		};

		/**Returns the index of the mexel.
		 */
		int getIndex() const throw ()
		{
			return m_index;
		};
		
		/**Sets the index of the mexel.
		 */
		void setIndex( unsigned int index ) throw ()
		{
			m_index = index;
		};
		
		/**Returns the shape parameter for the mexel.
		 */
		const MLVector & getAlpha() const throw ()
		{
			static MLVector sThinPlateAlpha(
				1.0 / 3.0,
				1.0 / 3.0,
				1.0 / 3.0);

			return sThinPlateAlpha;
		};
		
	}; // class MexelC
	
}; // namespace meshlib

#endif //MESHLIB_MEXELC_H
