/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "OFFMeshLoaderC.h"
#include "MeshC.h"
#include "MexelC.h"
#include "FaceC.h"

#include <iolib/LoaderExceptionC.h>

#include <fstream>

//#ifdef HAVE_NEW_IOSTREAMS
  #include <sstream>
  typedef std::istringstream my_isstream;
//#else
 // #include <strstream>
 // typedef istrstream my_isstream;
//#endif

using namespace meshlib;
using namespace std;
//using namespace iolib;
	

const OFFMeshLoaderC & OFFMeshLoaderC::getInstance()
{
	static OFFMeshLoaderC * pLoader = NULL;
	
	if( ! pLoader )
		pLoader = new OFFMeshLoaderC();
	return *pLoader;
}

OFFMeshLoaderC::~OFFMeshLoaderC() throw()
{
}

const std::string & OFFMeshLoaderC::getDescription() const throw()
{
	static string desc = "OFF";
	return desc;
}

bool OFFMeshLoaderC::accepts( const std::string & rFileName, const MeshC::ParametersC & rParams ) const throw()
{
	return accepts( rFileName );
}

bool OFFMeshLoaderC::accepts( const std::string & rFileName ) const throw()
{
	string buffer;	
	ifstream offFile( rFileName.c_str() );
	cout << offFile.good() << endl;
	getline( offFile, buffer );
	//ignore comments
	while ( offFile.good() && buffer.at(0) == '#')
		getline( offFile, buffer );
	if ( offFile.good() && buffer == "OFF" )
		return true;
	return false;
}

MeshC * OFFMeshLoaderC::load( const std::string & rFileName) const
	throw( iolib::ExceptionC )
{
	return load(
		rFileName,
		MeshC::ParametersC( MLVector(0,0,0) )
		);
}

/*void OFFMeshLoaderC::trim_string( std::string &s )
{
   s.erase( 0, s.find_first_not_of(" \t") );
   s.erase( s.find_last_not_of( " \t" ) + 1 );
}*/

MeshC * OFFMeshLoaderC::load( const std::string & rFileName, const MeshC::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	try {

	string buffer;		
	ifstream offFile( rFileName.c_str() );
	
	getline( offFile,  buffer );
	//ignore comments
	while ( offFile.good() && buffer.at(0) == '#')
		getline( offFile, buffer );

	if( ! offFile.good() || buffer != "OFF" )
	{
		offFile.close();
		throw iolib::LoaderExceptionC("File \'" + rFileName + "\' is not a valid OFF file");
	}
	
	getline( offFile,  buffer );
	unsigned int mexelCount;
	unsigned int faceCount;
	
	my_isstream is1(buffer);
	is1 >> mexelCount >> faceCount;
	
	// allocate a new mesh
	MeshC* pMesh = new MeshC();
	
	pMesh->setReferencePoint( rParams.getReferencePoint() );
		
	while ( mexelCount-- )
	{
		getline( offFile,  buffer );
		my_isstream is(buffer);
		MLFloat fX,fY,fZ;
		is >> fX >> fY >> fZ;
		pMesh->addMexel( MexelC( fX, fY, fZ ) );
	}
	
	while ( faceCount-- )
	{
		getline( offFile,  buffer );
//		char chDummy;
		my_isstream is(buffer);
		FaceC face;
		int pointCount;
		is >> pointCount;
		int index;
		is >> index;
		MexelC * pFirstMex = pMesh->getMexel( index );
		MexelC * pPrevMex = pFirstMex;
		MexelC * pMex = pFirstMex;
		face.addMexel( pMex );
		while ( --pointCount )
		{
			is >> index;
			if ( index < 0 || index >= (signed int)pMesh->getMexelCount() )
				throw iolib::LoaderExceptionC("Input mesh could not be loaded!");
				
			pMex = pMesh->getMexel( index );
			pPrevMex->addNeighbor(pMex);
			face.addMexel( pMex );
			pPrevMex = pMex;
		}
		pPrevMex->addNeighbor(pFirstMex);
		
		pMesh->addFace( face );
	}
	
	return pMesh;
	
	} catch ( ... ) {
		throw iolib::LoaderExceptionC("Error loading OFF file");
	}
}

void OFFMeshLoaderC::save( const MeshC * pProduct, const std::string & rFileName, const MeshC::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	ofstream offFile( rFileName.c_str() );

	offFile << "OFF\r\n";
	
	offFile
		<< pProduct->getMexelCount() << " "
		<< pProduct->getFaceCount() << " 0\r\n";


	MexelConstIterator itMex = pProduct->mexelBegin();
	while( itMex != pProduct->mexelEnd() )
	{
		const MLVector & rCoords = (*itMex)->getLocation();
		offFile
			<< rCoords(0) << " " << rCoords(1) << " " << rCoords(2)
			<< "\r\n";
		++itMex;
	}
		
	FaceConstIterator itFace = pProduct->faceBegin();
	while( itFace != pProduct->faceEnd() )
	{
		MexelConstIterator itMex2 = (*itFace)->mexelBegin();
		offFile << (*itFace)->getMexelCount() << " ";
		while( itMex2 != (*itFace)->mexelEnd() )
		{
			offFile << " " <<(*itMex2)->getIndex();
			++itMex2;
		}
		offFile << "\r\n";
		++itFace;
	}
	
	offFile << flush;
		
	offFile.close();
}

void OFFMeshLoaderC::save( const MeshC * pProduct, const std::string & rFileName) const
	throw( iolib::ExceptionC )
{
	save( pProduct, rFileName, MeshC::ParametersC("") );
}

OFFMeshLoaderC::OFFMeshLoaderC() throw()
{
}


