/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "VRMLMeshLoaderC.h"
#include "MeshC.h"
#include "MexelC.h"
#include "FaceC.h"

#include <iolib/LoaderExceptionC.h>

#include <fstream>

//#ifdef HAVE_NEW_IOSTREAMS
  #include <sstream>
  typedef std::istringstream my_isstream;
//#else
 // #include <strstream>
 // typedef istrstream my_isstream;
//#endif

using namespace meshlib;
using namespace std;
//using namespace iolib;

const VRMLMeshLoaderC & VRMLMeshLoaderC::getInstance()
{
	static VRMLMeshLoaderC * pLoader = NULL;
	
	if( ! pLoader )
		pLoader = new VRMLMeshLoaderC();
	return *pLoader;
}

VRMLMeshLoaderC::~VRMLMeshLoaderC() throw()
{
}

const std::string & VRMLMeshLoaderC::getDescription() const throw()
{
	static string desc = "VRML 1.0 (subset)";
	return desc;
}

bool VRMLMeshLoaderC::accepts( const std::string & rFileName, const MeshC::ParametersC & rParams ) const throw()
{
	return accepts( rFileName );
}

bool VRMLMeshLoaderC::accepts( const std::string & rFileName ) const throw()
{
	string buffer;	
	ifstream wrlFile( rFileName.c_str() );
	getline( wrlFile, buffer );
	trim_string(buffer);
	if( wrlFile.good() && buffer == "#VRML V1.0 ascii" )
		return true;
	return false;
}

MeshC * VRMLMeshLoaderC::load( const std::string & rFileName, const MeshC::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	return load(rFileName);
}

void VRMLMeshLoaderC::trim_string( std::string &s )
{
   s.erase( 0, s.find_first_not_of(" \t\r\n") );
   s.erase( s.find_last_not_of( " \t\r\n" ) + 1 );
}

MeshC * VRMLMeshLoaderC::load( const std::string & rFileName) const
	throw( iolib::ExceptionC )
{
	try {

	string buffer;
	ifstream wrlFile( rFileName.c_str() );
	
	getline( wrlFile,  buffer );
	trim_string(buffer);

	if( ! wrlFile.good() || buffer != "#VRML V1.0 ascii" )
	{
		wrlFile.close();
		throw iolib::LoaderExceptionC("File \'" + rFileName + "\' is not a valid VRML file");
	}
	
	//find separator node
	do {
		getline( wrlFile, buffer);
		trim_string(buffer);
	} while( wrlFile.good() && ( buffer.substr(0,string("Separator").size()) != "Separator" ));
	if ( ! wrlFile.good() )
		throw iolib::LoaderExceptionC("Error loading input mesh!");
	
	//find translation node
	do {
		getline( wrlFile, buffer);
		trim_string(buffer);
	} while( wrlFile.good() && ( buffer.substr(0,string("Translation").size()) != "Translation" ));
	if ( ! wrlFile.good() )
		throw iolib::LoaderExceptionC("Error loading input mesh!");

	// allocate a new mesh
	MeshC* pMesh = new MeshC();
	
	wrlFile >> buffer; //should be 'translation'
	float fX, fY, fZ;
	wrlFile >> fX >> fY >> fZ;
	
	pMesh->setReferencePoint( MLVector( fX, fY, fZ ) );
	
	//find coordinate3 node
	do {
		getline( wrlFile, buffer);
		trim_string(buffer);
	} while( wrlFile.good() && ( buffer.substr(0,string("Coordinate3").size()) != "Coordinate3" ));
	if ( ! wrlFile.good() )
		throw iolib::LoaderExceptionC("Error loading input mesh!");
	
	getline( wrlFile, buffer); // point [
	getline( wrlFile,  buffer );
	trim_string( buffer );	
	do
	{
		my_isstream is(buffer);
		is >> fX >> fY >> fZ;
		pMesh->addMexel( MexelC( fX, fY, fZ ) );
		getline( wrlFile,  buffer );
		trim_string( buffer );	
	} while( buffer != "]" );
	
	//find indexedfaceset node
	do {
		getline( wrlFile, buffer);
		trim_string(buffer);
	} while( wrlFile.good() && ( buffer.substr(0,string("IndexedFaceSet").size()) != "IndexedFaceSet" ));
	if ( ! wrlFile.good() )
		throw iolib::LoaderExceptionC("Error loading input mesh!");
	
	getline( wrlFile, buffer); // coordindex [
	getline( wrlFile,  buffer );
	trim_string( buffer );	
	do
	{
		char chDummy;
		my_isstream is(buffer);
		FaceC face;
		int index;
		is >> index;
		is >> chDummy; // comma
	
		if ( chDummy != ',')
			throw iolib::LoaderExceptionC("Error loading input mesh!");
		
		MexelC * pFirstMex = pMesh->getMexel( index );
		MexelC * pPrevMex = pFirstMex;
		MexelC * pMex = pFirstMex;
		face.addMexel( pMex );
		while( index != -1 )
		{
			is >> index;
			is >> chDummy; // comma

			if ( chDummy != ',')
				throw iolib::LoaderExceptionC("Error loading input mesh!");

			if(index != -1)
			{
				if ( index >= (signed int)pMesh->getMexelCount() )
					throw iolib::LoaderExceptionC("Error loading input mesh!");
					
				pMex = pMesh->getMexel( index );
				pPrevMex->addNeighbor(pMex);
				face.addMexel( pMex );
				pPrevMex = pMex;
			}
		}
		pPrevMex->addNeighbor(pFirstMex);
		
		pMesh->addFace( face );
		getline( wrlFile,  buffer );
		trim_string( buffer );	
	} while( buffer != "]" );
	
	return pMesh;
	
	} catch ( ... ) {
		throw iolib::LoaderExceptionC("Error loading VRML file");
	}
}

void VRMLMeshLoaderC::save( const MeshC * pProduct, const std::string & rFileName, const MeshC::ParametersC & rParams ) const
	throw( iolib::ExceptionC )
{
	ofstream wrlFile( rFileName.c_str() );

	wrlFile <<
		"#VRML V1.0 ascii\r\n"
		"\r\n";

	my_isstream is( rParams.getComment() );
	while( is.good() )
	{
		string s;
		getline( is,s );
		wrlFile << "# " << s << endl;
	}
	
	wrlFile <<
	  	"\r\n"
		"Separator {\r\n"
		"\r\n"
		"\tShapeHints {\r\n"
 		"\t\tvertexOrdering\tCOUNTERCLOCKWISE\r\n"
  		"\t\tfaceType\tCONVEX\r\n"
		"\t\tcreaseAngle\t0.5\r\n"
		"\t}\r\n"
		"\r\n"
		"\tMaterial {\r\n"
		"\t\tdiffuseColor\t.8 .8 .8\r\n"
		"\t}\r\n"
		"\r\n"
		"\tTranslation {\r\n"
		"\t\ttranslation\t";
		
	const MLVector & rPoint = pProduct->getReferencePoint();
	wrlFile << rPoint(0) << " " << rPoint(1) << " "  << rPoint(2) <<
		"\r\n"
		"\t}\r\n";

	wrlFile <<
		"\r\n"
		"\tCoordinate3 {\r\n"
		"\t\tpoint [\r\n";

	MexelConstIterator itMex = pProduct->mexelBegin();
	while( itMex != pProduct->mexelEnd() )
	{
		const MLVector & rCoords = (*itMex)->getLocation();
		wrlFile << "\t\t\t" << rCoords(0) << " " << rCoords(1) << " " << rCoords(2) << "," << endl;
		
		++itMex;
	}
		
	wrlFile <<
		"\t\t]\r\n"
		"\t}\r\n";

	wrlFile <<
		"\r\n"
		"\tIndexedFaceSet {\r\n"
		"\t\tcoordIndex [\r\n";

	FaceConstIterator itFace = pProduct->faceBegin();
	while( itFace != pProduct->faceEnd() )
	{
		MexelConstIterator itMex2 = (*itFace)->mexelBegin();
		wrlFile << "\t\t\t";
		while( itMex2 != (*itFace)->mexelEnd() )
		{
			wrlFile << (*itMex2)->getIndex() << ", ";
			++itMex2;
		}
		wrlFile << "-1," << endl;
		++itFace;
	}
	
	wrlFile <<
		"\t\t]\r\n"
		"\t}\r\n"
		"}\r\n" << flush;
		
	wrlFile.close();
}

void VRMLMeshLoaderC::save( const MeshC * pProduct, const std::string & rFileName) const
	throw( iolib::ExceptionC )
{
	save( pProduct, rFileName, MeshC::ParametersC("") );
}

VRMLMeshLoaderC::VRMLMeshLoaderC() throw()
{
}


