/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ParametersC.h"
#include <cassert>
//#include <stringstream>

//#ifdef HAVE_NEW_IOSTREAMS
  #include <sstream>
  typedef std::istringstream my_isstream;
//#else
//  #include <strstream>
//  typedef istrstream my_isstream;
//#endif

using namespace paramlib;
using namespace std;

ParametersC::ParametersC( const std::vector<DefinitionS> & definitions )
	throw( ParameterExceptionC )
{
	initialize( definitions, true );
}

void ParametersC::initialize(
	const std::vector<DefinitionS> & definitions,
	bool bInitDefaultValues ) throw( ParameterExceptionC )
{	
	if( definitions.size() == 0 )
		throw ParameterExceptionC("Parameter definition is empty.");
		
	std::vector<DefinitionS>::const_iterator itDefs = definitions.begin();
	while( itDefs != definitions.end() )
	{
		//check definition validity
		switch( itDefs->valueType )
		{
			case BOOL :
			case INT :
			case FLOAT :
			case VECTOR :
			case STRING :
			case NONE :
				break;
			default :
				throw ParameterExceptionC("Unknown parameter value type : " +
					itDefs->valueType );
		}
		
		if( ! itDefs->szNames || itDefs->szNames[0] == '\0' )
			throw ParameterExceptionC("Parameter name missing from definition.");
		
		//get key ID
		int iKeyID = itDefs->iKeyID;
		
		//fill definitions-by-ID map
		m_defsByID.insert( pair<int,DefinitionS>(iKeyID, *itDefs) );
		
		//fill values-by-ID map
		if( itDefs->szDefaultValue && bInitDefaultValues )
			m_valuesByID.insert(
				pair<int,string>( iKeyID, itDefs->szDefaultValue) );

		//fill IDs-by-name map
		//this piece of code parses the spaces of the szNames field
		unsigned int indexL = 0;
		unsigned int indexR = 0;
		string names( itDefs->szNames );
		while( indexL < names.length() )
		{
			indexR = names.find(" ",indexL);
			if( indexR == string::npos )
				indexR = names.length();
			m_IDsByName.insert( pair<string,int>(
				names.substr( indexL, indexR - indexL ), iKeyID ) );
			indexL = indexR + 1;
		}			

		//and aliases too
		if( itDefs->szAliases )
		{
			const char * pChar = itDefs->szAliases;
			while( *pChar != '\0' )
				m_IDsByName.insert( pair<string,int>(
					string( pChar++, 1 ), iKeyID ) );
		}
		
		++itDefs;
	}		
}

ParametersC::ParametersC(
	const std::vector<DefinitionS> & definitions,
	int argc,
	char * argv[] ) throw( ParameterExceptionC )
{
	initialize( definitions, false ); //init w/o default values.
	parseCommandLine( argc, argv );
}

ParametersC::ParametersC(
	const std::vector<DefinitionS> & definitions,
	std::istream & rStream )
	throw ( iolib::IOExceptionC, ParameterExceptionC )
{
	initialize( definitions, false ); //init w/o default values.
	parseConfig( rStream );
}

ParametersC::ParametersC( const ParametersC & rParams ) throw()
{
	operator=(rParams);
}

ParametersC::~ParametersC() throw()
{
	//nothing to free; all data is owned by caller.
}

const ParametersC ParametersC::operator+( const ParametersC & rParams ) const
	throw( ParameterExceptionC )
{
	ParametersC params(*this);
	
	std::map<int,string>::const_iterator itValue
		= rParams.m_valuesByID.begin();
		
	while ( itValue != rParams.m_valuesByID.end() )
	{
		int iKeyID = itValue->first;

		std::map<int,DefinitionS>::iterator itDef
			= params.m_defsByID.find(iKeyID);
			
		if ( itDef != m_defsByID.end() )
		{
			params.m_valuesByID.erase(iKeyID);
			params.m_valuesByID.insert(
				pair<int,string>(iKeyID, itValue->second ) );
			int i = 0;
			while ( itDef->second.overrides[i] )
			{
				if ( ! rParams.containsKey( itDef->second.overrides[i] ) )
					params.m_valuesByID.erase( itDef->second.overrides[i] );
				++i;
			}
		} else
			throw ParameterExceptionC("Incompatible parameter sets.");
			
		++itValue;
	}
	return params;
}

ParametersC & ParametersC::operator=( const ParametersC & rParams ) throw()
{
	if ( &rParams != this )
	{
		m_defsByID = rParams.m_defsByID;
		m_valuesByID = rParams.m_valuesByID;
		m_IDsByName = rParams.m_IDsByName;				
	}
	return *this;
}

std::string ParametersC::toString() const throw()
{
	ostringstream os;
	
	os << "Generated by dualsurfacemin v1.0\r\n\r\nParameter set:\r\n\r\n";

	map<int,string>::const_iterator itValue = m_valuesByID.begin();
	while(itValue != m_valuesByID.end())
	{
		string keyName( m_defsByID.find( itValue->first )->second.szNames );
		string::size_type index = keyName.find(" ");
		if( index != string::npos )
			keyName.erase( index );
		os << keyName << " = " << ( itValue->second ) << endl;
		++itValue;
	}
	return os.str();
}


ParametersC & ParametersC::operator+=( const ParametersC & rParams )
	throw ( ParameterExceptionC )
{
	return operator=( (*this) + rParams );
}

bool ParametersC::containsKey( int iKeyID ) const throw()
{
	return m_valuesByID.find( iKeyID ) != m_valuesByID.end();
}

const string & ParametersC::getString( int iKeyID ) const
	throw(ParameterExceptionC)
{
	return m_valuesByID.find( iKeyID )->second;
}

float ParametersC::getFloat( int iKeyID ) const
	throw(ParameterExceptionC)
{
	my_isstream is( m_valuesByID.find( iKeyID )->second );
	float f;
	is >> f;
	return f;
}

blitz::TinyVector<float,3> ParametersC::getVector( int iKeyID ) const
	throw(ParameterExceptionC)
{
	my_isstream is( m_valuesByID.find( iKeyID )->second );
	float fX,fY,fZ;
	is >> fX >> fY >> fZ;
	return blitz::TinyVector<float,3>(fX,fY,fZ);
}

int ParametersC::getInt( int iKeyID ) const throw(ParameterExceptionC)
{
	my_isstream is( m_valuesByID.find( iKeyID )->second );
	int i;
	is >> i;
	return i;
}

bool ParametersC::getBool( int iKeyID ) const throw(ParameterExceptionC)
{
	std::map<int,string>::const_iterator it = m_valuesByID.find(iKeyID);

	if ( it == m_valuesByID.end() )
		return false;
		
	if( it->second == "true")
		return true;
	else
		return false;
}

void ParametersC::addValue(
	const std::string & rKeyName,
	const std::string & rValue ) throw ( ParameterExceptionC )
{
	std::map<std::string,int>::const_iterator it
		= m_IDsByName.find( rKeyName );
		
	if( it != m_IDsByName.end() )
	{
		assert( m_IDsByName.count( rKeyName ) == 1 );

		unsigned int iKeyID = it->second;
		
		assert( iKeyID < m_defsByID.size() );
		
		ValueTypeE vt = m_defsByID[ iKeyID ].valueType;

		assert( vt == NONE || vt == BOOL || vt == INT ||
				vt ==FLOAT || vt ==STRING || vt == VECTOR );

		my_isstream is(rValue);
		switch( vt )
		{
			case NONE :
			{
				break;
			}
			case BOOL :
			{
				if(! (rValue == "true" || rValue == "false"))
					throw ParameterExceptionC("Invalid boolean value: "+rValue);
				break;
			}
			case INT :
			{
				int iDummy;
				is >> iDummy;
				if(is.fail())
					throw ParameterExceptionC("Invalid integer value: "+rValue);
				break;
			}
			case FLOAT :
			{
				float fDummy;
				is >> fDummy;
				if(is.fail())
					throw ParameterExceptionC("Invalid floating point value: "+rValue);				
				break;
			}
			case STRING :
			{
				break;
			}
			case VECTOR :
			{
				float x,y,z;
				is >> x >> y >> z;
				if(is.fail())
					throw ParameterExceptionC("Invalid vector value: "+rValue);								
				break;
			}
		}
		
		m_valuesByID.erase(iKeyID);
		m_valuesByID.insert( pair<int,string>(iKeyID, rValue) );
		
	} else {
		throw ParameterExceptionC("Unknown parameter key \'" + rKeyName + "\'");
	}			
}

void ParametersC::parseCommandLine( int argc, char * argv[] ) throw ( ParameterExceptionC )
{
	assert( argc >= 0 );
	assert( argv != NULL );
	
	for( int i = 1; i < argc; ++i )
	{
		string arg( argv[i] );
		string keyName;
		if( arg.substr(0,2) == "--" )
			keyName = arg.substr( 2 );
		else if( arg.substr(0,1) == "-" )
			keyName = arg.substr( 1 );
		else
			throw ParameterExceptionC("Command-line arguments must be prefixed with \'--\' or \'-\'.");
			
		std::map<std::string,int>::const_iterator it = m_IDsByName.find( keyName );
		if( it != m_IDsByName.end() )
		{
			assert( m_IDsByName.count( keyName ) == 1 );
			
			unsigned int iKeyID = it->second;
			
			assert( iKeyID < m_defsByID.size() );
			
			switch( m_defsByID[ iKeyID ].valueType )
			{
				case NONE :
				{
					m_valuesByID.insert( pair<int,string>(iKeyID, "") );
					break;
				}

				case BOOL :
				{
					if( i+1 < argc )
					{
						++i;
						string value = argv[ i ];
						// bugfix:
						if( value.substr(0,1) == "-" )
						{
							m_valuesByID.erase(iKeyID);
							m_valuesByID.insert( pair<int,string>(iKeyID, "true") );
							--i;
							break;
						}
							
						if(value == "true" || value == "1")
						{
							m_valuesByID.erase(iKeyID);
							m_valuesByID.insert( pair<int,string>(iKeyID, "true") );
						} else if(value == "false" || value == "0") {
							m_valuesByID.erase(iKeyID);
							m_valuesByID.insert( pair<int,string>(iKeyID, "false") );
						} else
							throw ParameterExceptionC( "Invalid boolean value : " + value );
					} else {
						m_valuesByID.erase(iKeyID);
						m_valuesByID.insert( pair<int,string>(iKeyID, "true") );
					}
					break;
				}

				case INT :
				{
					if( i+1 < argc )
					{
						++i;
						string value = argv[ i ];
						my_isstream is(value);
						int iValue;
						is >> iValue;
						if( is.fail() )
							throw ParameterExceptionC( "Invalid integer argument: " + value );
						m_valuesByID.erase(iKeyID);
						m_valuesByID.insert( pair<int,string>(iKeyID, value) );
					} else {
						throw ParameterExceptionC( "Integer argument expected after --" + keyName );
					}
					break;
				}

				case FLOAT :
				{
					if( i+1 < argc )
					{
						++i;
						string value = argv[ i ];
						my_isstream is(value);
						float fValue;
						is >> fValue;
						if( is.fail() )
							throw ParameterExceptionC( "Invalid floating point argument: " + value );
						m_valuesByID.erase(iKeyID);
						m_valuesByID.insert( pair<int,string>(iKeyID, value) );
					} else {
						throw ParameterExceptionC( "Floating point argument expected after --" + keyName );
					}
					break;
				}
				
				case VECTOR :
				{
					if( i+3 < argc )
					{
						string valueX = argv[ ++i ];
						string valueY = argv[ ++i ];
						string valueZ = argv[ ++i ];
						string value( valueX + " " + valueY + " " + valueZ );
						my_isstream is(value);
						float fX,fY,fZ;
						is >> fX >> fY>> fZ;
						if( is.fail() )
							throw ParameterExceptionC( "Invalid vector argument: " + value );
						m_valuesByID.erase(iKeyID);
						m_valuesByID.insert( pair<int,string>(iKeyID, value ));
					} else {
						throw ParameterExceptionC( "Vector argument expected after --" + keyName );
					}
					break;
				}

				case STRING :
				{
					if( i+1 < argc )
					{
						///@todo allow "quoted string arguments"
						string value = argv[ ++i ];
						m_valuesByID.erase(iKeyID);
						m_valuesByID.insert( pair<int,string>(iKeyID, value) );
						//m_valuesByID.find(iKeyID)->second = value;
					} else {
						throw ParameterExceptionC( "String argument expected after --" + keyName );
					}
					break;
				}

				default :
					throw ParameterExceptionC( "Unknown parameter value type : " + m_defsByID[iKeyID].valueType );
					break;
			}
			
			///@todo overrides???

		} else {
			throw ParameterExceptionC("Unknown command-line argument : " + keyName );
		}
	}
}

void ParametersC::trim_string( std::string &s )
{
   s.erase( 0, s.find_first_not_of(" \t") );
   s.erase( s.find_last_not_of( " \t" ) + 1 );
}

void ParametersC::parseConfig( std::istream & rStream ) throw ( ParameterExceptionC )
{
	while( rStream.good() )
	{
		string line;
		getline( rStream, line );
		trim_string( line );
		if( ! ( line.substr(0,1) == "#" || line == "" ) )
		{
			my_isstream is( line );
			string keyName;
			is >> keyName;
			string buffer;
			is >> buffer;
			trim_string( buffer );
			if( buffer != "=" )
				throw ParameterExceptionC("Error parsing config file; \'=\' expected");
			getline( is, buffer );
			size_t index = buffer.find_first_of("#");
			if( index != string::npos )
				buffer.erase( index );
			trim_string( buffer );
			addValue( keyName, buffer );
		}		
	}
}


