/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef PARAMLIB_PARAMETERSC_H
#define PARAMLIB_PARAMETERSC_H

#include "ParameterExceptionC.h"
#include <iolib/IOExceptionC.h>
#include <blitz/tinyvec.h>
#include <map>
#include <vector>
#include <string>
#include <istream>

namespace paramlib
{
	/**Represents a parameter set of key/value pairs.
	 */
	class ParametersC
	{
	public:
		
		/**Type of the parameter.
		 */
		enum ValueTypeE
		{
			NONE = 0,
			BOOL,
			INT,
			FLOAT,
			VECTOR,			
			STRING
		};
	
		/**A structure for a parameter definition.
		 * The user defines her own set of acceptable parameters and their
		 * relations by passing a vector of DefinitionSs to the
		 * ParametersC constructor.
		 */
		class DefinitionS
		{
		public:
			/// ID for this parameter key.
 			int iKeyID;

 			/// Type of the parameter.
			ValueTypeE valueType;

			/// Names separated by spaces.
			const char * szNames;

			/// Aliases.
			const char * szAliases;

			/** Description of the value, e.g. "true | false".
			 * Can be NULL.
			 */			
			const char * szValueDescription;
			
			///Short description of the parameter.
			const char * szDescription; 
			
			///Default value.
			const char * szDefaultValue;

			/**Parameters defined here will be overridden by this
			 * parameter in the +-operation.
			 */
			int overrides[5]; 
		};
		
		/**Construct a ParametersC object and initializes all parameters
		 * to their default values.
		 */
		ParametersC( const std::vector<DefinitionS> & definitions )
			throw ( ParameterExceptionC );

		/**Constructs a ParametersC object and initializes all
		 * parameters to values from the command line.
		 */
		ParametersC(
			const std::vector<DefinitionS> & definitions,
			int argc,
			char * argv[] ) throw ( ParameterExceptionC );
			
		/**Constructs a ParametersC object and initializes all
		 * parameters to values from a config file.
		 */
		ParametersC(
			const std::vector<DefinitionS> & definitions,
			std::istream & rStream )
			throw ( iolib::IOExceptionC, ParameterExceptionC );
			
		/**Copy constructor.
		 */
		ParametersC( const ParametersC & rParams ) throw();
		
		/**A non-virtual destructor.
		 */
		~ParametersC() throw();
		
		/**Merge operation.
		 * Parametes will be overridden by those in @a rParams.
		 */
		const ParametersC operator+( const ParametersC & rParams ) const
			throw ( ParameterExceptionC );
		
		/**Merge operation.
		 * Parametes will be overridden by those in @a rParams.
		 */
		ParametersC & operator+=( const ParametersC & rParams )
			throw ( ParameterExceptionC );
			
		/**Assignment.
		 */
		ParametersC & operator=( const ParametersC & rParams ) throw();

		/**Returns the parameter set as a string.
		 */
		std::string toString() const throw();
		
		/**Returns true if the key is present.
		 */
		bool containsKey( int iKeyID ) const throw();
		
		/**Returns a string value by key ID.
		 */
		const std::string & getString( int iKeyID ) const
			throw ( ParameterExceptionC );
			
		/**Returns a float value by key ID.
		 */
		float getFloat( int iKeyID ) const throw ( ParameterExceptionC );
		
		/**Returns a vector value by key ID.
		 */		
		blitz::TinyVector<float,3> getVector( int iKeyID ) const
			throw ( ParameterExceptionC );

		/**Returns an integer value by key ID.
		 */		
		int getInt( int iKeyID ) const throw ( ParameterExceptionC );
		
		/**Returns a boolean value by key ID.
		 */
		bool getBool( int iKeyID ) const throw ( ParameterExceptionC );

	protected:

		void initialize(
			const std::vector<DefinitionS> & definitions,
			bool bInitDefaultValues = false  ) throw ( ParameterExceptionC );

		void parseCommandLine( int argc, char * argv[] )
			throw ( ParameterExceptionC );
			
		void parseConfig( std::istream & rStream )
			throw ( ParameterExceptionC );
			
		void addValue(
			const std::string & rKeyName,
			const std::string & rValue ) throw ( ParameterExceptionC );
			
		static void trim_string( std::string& );
		
		std::map<int, std::string> m_valuesByID;
		std::map<int, DefinitionS> m_defsByID;
		std::map<std::string, int> m_IDsByName;
	
	}; //class ParametersC
	
}; // namespace paramlib

#endif //PARAMLIB_PARAMETERSC_H
